import os
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Callable, Dict, List, Optional

import matplotlib.pyplot as plt
import pandas as pd
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import MinMaxScaler, StandardScaler

import nemo_bo.utils.logger as logging_nemo
import nemo_bo.utils.perf_metrics as pm
import nemo_bo.utils.plotter as plotter
from nemo_bo.models.base.available_models import create_predictor_list
from nemo_bo.models.base.base_model import Base_Model
from nemo_bo.opt.variables import VariablesList
from nemo_bo.utils.data_proc import *
from nemo_bo.utils.transformations import Transformations

try:
    logging_nemo.logging_path
    logger = logging_nemo.logging_nemo_child(os.path.basename(__file__))
except AttributeError:
    logger = logging_nemo.logging_nemo_master(os.path.basename(__file__))


@dataclass(kw_only=True)
class Objective:
    name: str
    obj_max_bool: bool
    lower_bound: int | float
    upper_bound: int | float
    transformation_type: Optional[str] = None
    units: str = ""

    def transform(self, Y: np.ndarray) -> np.ndarray:
        if self.transformation_type is None:
            raise AttributeError(
                "Please either: 1. Fit the model first, or 2. Set transformation_type = 'none', 'normalisation', or 'standarisation' when creating the objective instance"
            )

        if self.transformation_type == "none":
            return Y
        self.transformations = Transformations()
        if self.transformation_type == "normalisation":
            return self.transformations.minmaxscaler(Y)
        elif self.transformation_type == "standardisation":
            return self.transformations.standardscaler(Y)

    def transform_only(self, Y: np.ndarray) -> np.ndarray:
        if self.transformation_type is None:
            raise AttributeError(
                "Please either: 1. Fit the model first, or 2. Set transformation_type = 'none', 'normalisation', or 'standarisation' when creating the objective instance"
            )

        if self.transformation_type == "none":
            return Y
        if self.transformation_type == "normalisation":
            return self.transformations.minmaxscaler_transform_only(Y)
        elif self.transformation_type == "standardisation":
            return self.transformations.standardscaler_transform_only(Y)

    def inverse_transform(self, Y_transform: np.ndarray) -> np.ndarray:
        if self.transformation_type == "none":
            return Y_transform
        if isinstance(self.transformations.scaler, MinMaxScaler):
            return self.transformations.inverse_minmaxscaler(Y_transform)
        elif isinstance(self.transformations.scaler, StandardScaler):
            return self.transformations.inverse_standardscaler(Y_transform)


@dataclass(kw_only=True)
class RegressionObjective(Objective):
    predictor_type: Optional[str | Base_Model | List[str | Base_Model]] = None
    predictor_params_dict: Optional[Dict[str, Dict[str, Callable]]] = None
    gp_kernel_choices: Optional[list] = None
    model_params: Optional[Dict[str, Any]] = None

    def __post_init__(self):
        self.input_predictor_type = self.predictor_type = create_predictor_list(self.predictor_type)
        self.input_predictor_params_dict = self.predictor_params_dict

    def fit_regressor(
        self, X: np.ndarray, Y: np.ndarray, variables: VariablesList, test_ratio: Optional[float] = 0.2, **kwargs
    ) -> None:
        X, Y = remove_nan(X, Y)

        self.obj_function = self.predictor_type(variables, self)
        self.obj_function.fit(X, Y, test_ratio=test_ratio, **self.model_params, **kwargs)

        # Calculate permutation feature importance
        (
            self.permutation_feature_importance,
            _,
        ) = self.obj_function.permutation_feature_importance_continuous(X, Y)

    def predict(self, X: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        Y_pred, Y_pred_stddev = self.obj_function.predict(X)

        return Y_pred.astype("float"), Y_pred_stddev.astype("float")

    def model_and_hyperparameter_opt(
        self,
        X: np.ndarray,
        Y: np.ndarray,
        variables: VariablesList,
        model_search_bool: bool,
        test_ratio: Optional[float] = 0.2,
        **kwargs,
    ) -> None:
        X, Y = remove_nan(X, Y)

        if model_search_bool:
            test_loss_list = []
            model_list = []
            model_params_list = []

            # Splits the dataset to create training and test datasets
            if kwargs.get("sort_before_split", True):
                (
                    X_train,
                    X_test,
                    Y_train,
                    Y_test,
                ) = sort_train_test_split_shuffle(X, Y, test_ratio=test_ratio, seed=1)
            else:
                X_train, X_test, Y_train, Y_test = train_test_split(X, Y, test_size=test_ratio, random_state=1)

            # Performs hyperopt for each predictor type passed into this function and appends each respective best model and hyperparameters
            for predictor_class in self.input_predictor_type:
                predictor_instance = predictor_class(variables, self)
                model, model_params = predictor_instance.hyperparam_opt(
                    X_train, Y_train, test_ratio, predictor_params_dict=self.input_predictor_params_dict, **kwargs
                )
                model_list.append(model)
                model_params_list.append(model_params)

            # Determines the test set prediction accuracy using each optimised predictor type
            for m in model_list:
                Y_test_pred, _ = m.predict(X_test)
                dict = pm.all_performance_metrics(Y_test, Y_test_pred)
                test_loss_list.append(dict["RMSE"])

            # Identifies which predictor type had the most accurate test set prediction
            best_index = test_loss_list.index(min(test_loss_list))
            self.predictor_type = self.input_predictor_type[best_index]
            self.model_params = model_params_list[best_index]

        elif self.obj_function.always_hyperparam_opt:
            # No model search; only hyperopt for the current best predictor type (identified previously)
            predictor_instance = self.predictor_type(variables, self)
            _, self.model_params = predictor_instance.hyperparam_opt(
                X, Y, test_ratio, predictor_params_dict=self.input_predictor_params_dict, **kwargs
            )

        # Re-fits the model but with the complete dataset
        self.fit_regressor(X, Y, variables, test_ratio=test_ratio, **kwargs)

    def parity_plot(self) -> None:

        if self.obj_function.include_validation:
            # Produces a 2D scatter plot showing the parity plot data points, and the x = y line as a line plot
            train_paritydata = self.obj_function.performance_metrics["Train Parity Data"]
            train_rmse = self.obj_function.performance_metrics["Train RMSE"]
            train_r2 = self.obj_function.performance_metrics["Train r2"]
            val_paritydata = self.obj_function.performance_metrics["Validation Parity Data"]
            val_rmse = self.obj_function.performance_metrics["Validation RMSE"]
            val_r2 = self.obj_function.performance_metrics["Validation r2"]

            scatter_parity_data = [train_paritydata, val_paritydata]
            error = [
                self.obj_function.Y_train_error,
                self.obj_function.Y_val_error,
            ]
            max_parity = np.amax(np.vstack([train_paritydata, val_paritydata]))
            min_parity = np.amin(np.vstack([train_paritydata, val_paritydata]))
            # Array for the x = y line in the parity plot
            x_equals_y = [
                [min_parity, min_parity],
                [max_parity, max_parity],
            ]
            scatter_legend = (
                f"Train ({train_paritydata.shape[0]} experiments) (95% CI)",
                f"Test ({val_paritydata.shape[0]} experiments) (95% CI)",
            )
            plot_title = f"{self.predictor_type} Model {self.name} Parity Plot\nTrain RMSE = {round(train_rmse.astype('float'), 2)} {self.units}, Train r2 = {round(train_r2, 2)}, \nTest RMSE = {round(val_rmse.astype('float'), 2)} {self.units}, Test r2 = {round(val_r2, 2)}"
            path = os.path.join(os.getcwd(), "ML Models", f"{self.obj_function.name}", "Parity Plots")
            if not os.path.exists(path):
                os.makedirs(path)
            output_file = os.path.join(
                path,
                rf"{self.name} {self.obj_function.name} Model Parity Plot, {datetime.now().strftime('%d-%m-%Y, %H-%M-%S')}",
            )

        else:
            train_paritydata = self.obj_function.performance_metrics["Parity Data"]
            train_rmse = self.obj_function.performance_metrics["RMSE"]
            train_r2 = self.obj_function.performance_metrics["r2"]
            scatter_parity_data = [train_paritydata]
            error = [self.obj_function.Y_pred_error]
            max_parity = np.amax(np.vstack([train_paritydata]))
            min_parity = np.amin(np.vstack([train_paritydata]))
            # Array for the x = y line in the parity plot
            x_equals_y = [
                [min_parity, min_parity],
                [max_parity, max_parity],
            ]
            scatter_legend = f"Train ({train_paritydata.shape[0]} experiments) (95% CI)"
            plot_title = f"{self.predictor_type} Model {self.name} Parity Plot\nRMSE = {round(train_rmse.astype('float'), 2)} {self.units}, r2 = {round(train_r2, 2)}"
            path = os.path.join(os.getcwd(), "ML Models", f"{self.obj_function.name}", "Parity Plots")
            if not os.path.exists(path):
                os.makedirs(path)
            output_file = os.path.join(
                path,
                rf"{self.name} {self.obj_function.name} Model Parity Plot, {datetime.now().strftime('%d-%m-%Y, %H-%M-%S')}",
            )

        plotter.plot(
            plot_dim="2D",
            scatter_data=scatter_parity_data,
            error=error,
            line_data=x_equals_y,
            scatter_legend=scatter_legend,
            line_legend="x = y",
            xlabel=f"Actual {self.units}",
            ylabel=f"Predicted {self.units}",
            plottitle=plot_title,
            output_file=output_file,
        )
        plt.close()

    def partial_dependence_plot(self, X: np.ndarray) -> Dict[str, pd.DataFrame]:
        partial_dependence_plot_dict = self.obj_function.partial_dependence_plot(X)

        return partial_dependence_plot_dict

    def cv(self, X: np.ndarray, Y: np.ndarray, test_ratio: float = 0.2, **kwargs) -> Dict[str, Any]:
        X, Y = remove_nan(X, Y)

        cv_results = self.obj_function.cv(
            X,
            Y,
            self.model_params,
            test_ratio=test_ratio,
            **kwargs,
        )

        return cv_results

    def y_scrambling_cv(
        self,
        X: np.ndarray,
        Y: np.ndarray,
        test_ratio: float = 0.2,
        plot_parity: bool = True,
        inc_error_bars: bool = False,
        **kwargs,
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        X, Y = remove_nan(X, Y)

        cv_results, cv_results_y_scrambling = self.obj_function.y_scrambling_cv(
            X,
            Y,
            self.model_params,
            test_ratio=test_ratio,
            plot_parity=plot_parity,
            inc_error_bars=inc_error_bars,
            **kwargs,
        )

        return cv_results, cv_results_y_scrambling


class ObjectiveFunction:
    def __init__(self, obj_function_data: Any):
        self.obj_function_data = obj_function_data

    def evaluate(self, X: np.ndarray, *args, **kwargs) -> None:
        pass


# @dataclass(init=False)
@dataclass(kw_only=True)
class CalculableObjective(Objective):
    # def __init__(
    #     self,
    #     name: str,
    #     obj_max_bool: bool,
    #     lower_bound: int | float,
    #     upper_bound: int | float,
    #     obj_function: ObjectiveFunction,
    #     units: str = "",
    # ):
    #     self.obj_function = obj_function
    #     super().__init__(name, obj_max_bool, lower_bound, upper_bound, units)
    obj_function: ObjectiveFunction

    def calculate(self, X: np.ndarray, *args, **kwargs) -> Tuple[np.ndarray, np.ndarray]:
        Y = self.obj_function.evaluate(X, *args, **kwargs)
        Y_stddedv = np.zeros_like(Y)

        return Y.astype("float"), Y_stddedv


@dataclass
class ObjectivesList:
    objectives: List[Objective]

    def __post_init__(self):
        self.n_obj = len(self.objectives)
        self.names = [obj.name for obj in self.objectives]
        self.units = [obj.units for obj in self.objectives]
        self.max_bool_dict = {obj.name: obj.obj_max_bool for obj in self.objectives}
        self.bounds = np.array(
            [
                [obj.lower_bound for obj in self.objectives],
                [obj.upper_bound for obj in self.objectives],
            ]
        ).T

    def fit(
        self,
        X: np.ndarray,
        Y: np.ndarray,
        variables: VariablesList,
        model_search_bool: Optional[bool] = None,
        **kwargs,
    ) -> None:
        if model_search_bool is None:
            model_search_bool = kwargs.get("model_search_bool", False)

        # Fit the ML regression models
        for obj_index, obj in enumerate(self.objectives):
            if isinstance(obj, RegressionObjective):
                obj.model_and_hyperparameter_opt(X, Y[:, obj_index], variables, model_search_bool, **kwargs)

    def evaluate(self, X: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        Y = np.zeros((X.shape[0], len(self.objectives)), dtype=np.float32)
        Y_stddev = np.zeros_like(Y)
        for obj_index, obj in enumerate(self.objectives):
            if isinstance(obj, RegressionObjective):
                Y[:, obj_index], Y_stddev[:, obj_index] = obj.predict(X)
            elif isinstance(obj, CalculableObjective):
                Y[:, obj_index], Y_stddev[:, obj_index] = obj.calculate(X)

        return Y.astype("float"), Y_stddev.astype("float")

    def parity_plot(self) -> None:
        for obj in self.objectives:
            if isinstance(obj, RegressionObjective):
                obj.parity_plot()

    def partial_dependence_plot(self, X: np.ndarray) -> Dict[str, Dict[str, pd.DataFrame]]:
        partial_dependence_plots_dict = {}
        for obj in self.objectives:
            if isinstance(obj, RegressionObjective):
                partial_dependence_plots_dict[obj.name] = obj.partial_dependence_plot(X)

        return partial_dependence_plots_dict

    def cv(self, X: np.ndarray, Y: np.ndarray, test_ratio: float = 0.2, **kwargs) -> Dict[str, Dict[str, Any]]:
        cv_dict = {}
        for obj_index, obj in enumerate(self.objectives):
            if isinstance(obj, RegressionObjective):
                cv_dict[obj.name] = obj.cv(X, Y[:, obj_index], test_ratio=test_ratio, **kwargs)

        return cv_dict

    def y_scrambling_cv(
        self,
        X: np.ndarray,
        Y: np.ndarray,
        test_ratio: float = 0.2,
        plot_parity: bool = True,
        inc_error_bars: bool = False,
        **kwargs,
    ) -> Tuple[Dict[str, Dict[str, Any]], Dict[str, Dict[str, Any]]]:
        cv_dict, cv_y_scrambling_dict = {}, {}
        for obj_index, obj in enumerate(self.objectives):
            if isinstance(obj, RegressionObjective):
                cv_dict[obj.name], cv_y_scrambling_dict[obj.name] = obj.y_scrambling_cv(
                    X,
                    Y[:, obj_index],
                    test_ratio=test_ratio,
                    plot_parity=plot_parity,
                    inc_error_bars=inc_error_bars,
                    **kwargs,
                )

        return cv_dict, cv_y_scrambling_dict



import os
from dataclasses import dataclass
from typing import Dict, List

import numpy as np

import nemo_bo.utils.logger as logging_nemo

try:
    logging_nemo.logging_path
    logger = logging_nemo.logging_nemo_child(os.path.basename(__file__))
except AttributeError:
    logger = logging_nemo.logging_nemo_master(os.path.basename(__file__))


@dataclass
class AcquisitionFunction:
    def build_ref_point(self, max_bool_dict: Dict[str, bool]) -> List[float]:
        ref_point = []
        hv_obj_minmax = np.array([[-0.01, 1.01] for _ in enumerate(max_bool_dict)])
        for obj_index, obj in enumerate(max_bool_dict):
            if max_bool_dict[obj]:
                ref_point.append(hv_obj_minmax[obj_index][0])
            else:
                ref_point.append(hv_obj_minmax[obj_index][1] * -1)

        return ref_point

    def Y_norm_minmax_transform(self, Y, bounds: np.ndarray, max_bool_dict: Dict[str, bool]) -> np.ndarray:
        # Transforms Y using min_max normalisation
        sign_adjusted_Y = np.zeros_like(Y)
        for obj_index, obj in enumerate(max_bool_dict):
            sign_adjusted_Y[:, obj_index] = (Y[:, obj_index] - bounds[obj_index, 0]) / (
                bounds[obj_index, 1] - bounds[obj_index, 0]
            )

            # To determine the pareto front, maximising objective values are left unchanged and minimising objective values are multiplied by -1
            if not max_bool_dict[obj]:
                sign_adjusted_Y[:, obj_index] *= -1

        return sign_adjusted_Y

    def remove_nan_rows(self, Y: np.ndarray) -> np.ndarray:
        # Acquisition functions do not allow for irregular arrays of objective values. Any rows that contains missing values (NaN) are removed
        if Y.ndim == 1:
            return Y[~np.isnan(Y)]
        else:
            return Y[~np.isnan(Y).any(axis=1)]



import sys
from typing import Any, Dict

from keras.models import Model


class SaveModelCheckPoint:
    """
    Class that enables model saving based on both the train loss and validation loss values
    """

    def __init__(self, model: Model, filepath: str):
        """
        Parameters
        ----------
        model : KerasRegressor type object that contains the neural network being trained
        filepath : str type of the file path to save the .h5 file that stores the model parameters

        """
        self.model = model
        self.filepath = filepath
        self.reset()

    def reset(self) -> None:
        """
        Function to reset the variables that are monitored during the check pointing process. This function can be called at the end of model training to allow multiple models to be trained using the same class instantiation of SaveModelCheckPoint
        """
        self.best_epoch = 0
        self.best_loss = sys.float_info.max
        self.best_val_loss = sys.float_info.max

    def save_weights(self, epoch: int, logs: Dict[str, Any]) -> None:
        """
        Function called during the model training procedure when a LambdaCallback is used that allows loss monitoring and model savings.

        Parameters
        ----------
        epoch : int type of the current training epoch
        logs : dict type containing loss information for the current training epoch

        """
        loss = logs["loss"]
        val_loss = logs["val_loss"]

        if epoch == 0:
            for _ in range(10):
                try:
                    self.model.save_weights(self.filepath)  # use TF save_weights method to save best model.
                    break
                except OSError as e:
                    print(e)

        elif loss < self.best_loss and val_loss < self.best_val_loss:
            self.best_loss = loss
            self.best_val_loss = val_loss
            self.best_epoch = epoch
            print(
                f"This epoch's loss ({loss}) and validation loss ({val_loss}) are both the best so far. Saving the model weights.\n"
            )
            for _ in range(10):
                try:
                    self.model.save_weights(self.filepath)  # use TF save_weights method to save best model.
                    break
                except OSError as e:
                    print(e)

        else:
            print(f"Number of epochs since the weights were last saved = {epoch - self.best_epoch}.\n")



from __future__ import annotations

import copy
import os
from functools import partial
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple

import gpytorch
import numpy as np
import torch
from gpytorch.distributions import MultivariateNormal
from gpytorch.kernels import AdditiveStructureKernel, Kernel, MaternKernel, ScaleKernel
from gpytorch.likelihoods import GaussianLikelihood, Likelihood
from gpytorch.means import ConstantMean
from gpytorch.models import ExactGP
from hyperopt import STATUS_OK, Trials, fmin, hp, tpe
from hyperopt.pyll.base import Literal
from torch import Tensor

import nemo_bo.utils.logger as logging_nemo
import nemo_bo.utils.perf_metrics as pm
from nemo_bo.models.base.base_model import Base_Model

if TYPE_CHECKING:
    from nemo_bo.opt.objectives import RegressionObjective
    from nemo_bo.opt.variables import VariablesList

try:
    logging_nemo.logging_path
    logger = logging_nemo.logging_nemo_child(os.path.basename(__file__))
except AttributeError:
    logger = logging_nemo.logging_nemo_master(os.path.basename(__file__))


class ExactGPModel(ExactGP):
    """
    Class for defining the GP model for regression (inherits from ExactGP in GPyTorch)
    """

    def __init__(self, trainX: Tensor, trainY: Tensor, likelihood: Likelihood, kernel: Kernel):
        super(ExactGPModel, self).__init__(trainX, trainY, likelihood)
        self.mean_module = ConstantMean()
        self.covar_module = kernel

    def forward(self, x: Tensor) -> MultivariateNormal:
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return MultivariateNormal(mean_x, covar_x)


class GPModel(Base_Model):

    """
    Class to instantiate for fitting GP models
    """

    def __init__(
        self, variables: VariablesList, objective: RegressionObjective, always_hyperparam_opt: bool = True, **kwargs
    ):
        super().__init__(variables, objective, always_hyperparam_opt)
        self.default_X_transform_type = "normalisation"
        self.default_Y_transform_type = "standardisation"
        self.include_validation = False
        self.name = "gp"

    def fit(
        self, X: np.ndarray, Y: np.ndarray, test_ratio: Optional[float] = None, sort_before_split: bool = None, **params
    ) -> None:
        """
        Function for fitting the model

        Parameters
        ----------
        plot_parity : bool type for generating the parity plot for the model fitting performance
        params : dict type containing keyword arguments that can be generated by the HyperOpt package for the model being fitted or chosen manually passing a dictionary
            The keyword arguments that can be used in this function are limited to the type of kernel and the number of training iterations

        """
        del test_ratio
        del sort_before_split
        if self.variables.num_cat_var > 0:
            self.X_train = self.variables.categorical_transform(X).astype("float")

        self.X_train, self.Y_train = self.transform_by_predictor_type(self.X_train, Y=Y)

        self.fit_model(params)

        self.Y_pred, self.Y_pred_stddev = self.predict(X)
        self.performance_metrics = pm.all_performance_metrics(Y, self.Y_pred)

        self.Y_pred_error = 1.96 * self.Y_pred_stddev

    def fit_model(self, params: Dict[str, Any]) -> None:

        X_tensor = torch.tensor(
            self.X_train.astype(np.float32),
            device=torch.device("cuda:0" if torch.cuda.is_available() else "cpu"),
        )
        Y_tensor = torch.tensor(
            self.Y_train.astype(np.float32),
            device=torch.device("cuda:0" if torch.cuda.is_available() else "cpu"),
        ).flatten()

        self.n_iterations = int(params.get("n_iterations", 50))
        self.kernel = params.get("kernel", ScaleKernel(MaternKernel(nu=2.5)))

        self.likelihood = GaussianLikelihood()
        self.model = ExactGPModel(X_tensor, Y_tensor, self.likelihood, copy.deepcopy(self.kernel))
        if torch.cuda.is_available():
            self.likelihood = self.likelihood.to("cuda")
            self.model = self.model.to("cuda")

        self.model.train()
        self.likelihood.train()

        self.optimizer = torch.optim.Adam(self.model.parameters())

        mll = gpytorch.mlls.ExactMarginalLogLikelihood(self.likelihood, self.model)

        for i in range(self.n_iterations):
            self.optimizer.zero_grad()
            output = self.model(X_tensor)
            mll_loss = -mll(output, Y_tensor).sum()
            mll_loss.backward()
            self.optimizer.step()

        self.model.eval()
        self.likelihood.eval()

    def predict(self, X: np.ndarray, X_transform: bool = True) -> Tuple[np.ndarray, np.ndarray]:
        # The functions below require the data to be a 2D array
        if X.ndim == 1:
            X = X.reshape(1, -1)

        if X_transform:
            if self.variables.num_cat_var > 0:
                X = self.variables.categorical_transform(X).astype("float")
            X = self.transform_only_by_predictor_type(X)

        # Split to reduce memory requirements
        split_size = 50000
        X_split = np.split(X, np.array((range(1, int(X.shape[0] / split_size) + 1))) * split_size)
        observed_Y_pred_list = []
        for X_split_each in X_split:
            if X_split_each.shape[0] == 0:
                continue
            with torch.no_grad():
                observed_Y_pred = self.likelihood(
                    self.model(
                        torch.tensor(
                            X_split_each.astype(np.float32),
                            device=torch.device("cuda:0" if torch.cuda.is_available() else "cpu"),
                        )
                    )
                )
            observed_Y_pred_list.append(observed_Y_pred)

        Y_pred = np.hstack([observed_Y_pred.mean.cpu().detach().numpy() for observed_Y_pred in observed_Y_pred_list])
        Y_pred_upper = np.hstack(
            [observed_Y_pred.confidence_region()[1].cpu().detach().numpy() for observed_Y_pred in observed_Y_pred_list]
        )

        Y_pred = self.objective.inverse_transform(Y_pred)
        Y_pred_upper = self.objective.inverse_transform(Y_pred_upper)
        Y_pred_stddev = (np.subtract(Y_pred_upper, Y_pred)) / 1.96

        return Y_pred.flatten(), Y_pred_stddev.flatten()

    @staticmethod
    def default_params(n_var, gp_kernel_choices: Optional[List[Kernel]] = None) -> Dict[str, Any]:
        if gp_kernel_choices is None:
            gp_kernel_choices = [
                AdditiveStructureKernel(
                    base_kernel=ScaleKernel(MaternKernel(ard_num_dims=n_var, nu=0.5)),
                    num_dims=n_var,
                ),
                AdditiveStructureKernel(
                    base_kernel=ScaleKernel(MaternKernel(ard_num_dims=n_var, nu=1.5)),
                    num_dims=n_var,
                ),
                AdditiveStructureKernel(
                    base_kernel=ScaleKernel(MaternKernel(ard_num_dims=n_var, nu=2.5)),
                    num_dims=n_var,
                ),
                ScaleKernel(MaternKernel(nu=0.5, ard_num_dims=n_var)),
                ScaleKernel(MaternKernel(nu=1.5, ard_num_dims=n_var)),
                ScaleKernel(MaternKernel(nu=2.5, ard_num_dims=n_var)),
            ]
        return {
            "n_iterations": hp.quniform("n_iterations", 5, 500, 1),
            "kernel": hp.choice("kernel", gp_kernel_choices),
        }

    def cv(
        self,
        X: np.ndarray,
        Y: np.ndarray,
        model_params: Dict[str, Any],
        test_ratio: float = 0.2,
        sort_before_split: Optional[bool] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        return self.cv_train_test(X, Y, model_params, test_ratio=test_ratio, **kwargs)

    def hyperparam_opt(
        self,
        X: np.ndarray,
        Y: np.ndarray,
        test_ratio: float,
        predictor_params_dict: Optional[Dict[str, Dict[str, Callable]]] = None,
        gp_kernel_choices: Optional[list] = None,
        **kwargs,
    ) -> Tuple[float, Dict[str, Any]]:

        if predictor_params_dict is None:
            predictor_params_dict = self.default_params(self.variables.n_var, gp_kernel_choices)
        else:
            predictor_params_dict = predictor_params_dict[self.name]

        max_evals = kwargs.get("max_evals", 1)
        # max_evals = kwargs.get("max_evals", 40)

        def func(X: np.ndarray, Y: np.ndarray, model_params: Dict[str, Any]) -> Dict[str, Any]:
            cv_results = self.cv(X, Y, model_params, test_ratio=test_ratio, **kwargs)

            return {
                "loss": cv_results["Mean Test RMSE"],
                "status": STATUS_OK,
            }

        trials = Trials()
        model_params = fmin(
            fn=partial(func, X, Y),
            space=predictor_params_dict,
            algo=tpe.suggest,
            max_evals=max_evals,
            trials=trials,
        )

        # HyperOpt returns an integer for the best GP kernel. This links back to the actual kernel.
        gp_kernel_choices = [i.obj for i in predictor_params_dict["kernel"].pos_args if isinstance(i, Literal)]
        model_params["kernel"] = gp_kernel_choices[model_params["kernel"]]

        model = self.new_instance(self.__class__, **kwargs)
        model.fit(X, Y, test_ratio=test_ratio, **self.model_params, **kwargs)

        logger.info(f"Completed hyperparameter opt")
        return model, model_params



import os
from dataclasses import dataclass
from itertools import combinations
from typing import Optional, Tuple

import numpy as np
import torch
from botorch.utils.multi_objective.hypervolume import Hypervolume
from botorch.utils.multi_objective.pareto import is_non_dominated
from pymoo.core.result import Result
from pymoo.factory import get_reference_directions, get_termination
from pymoo.optimize import minimize
from scipy import spatial

import nemo_bo.utils.logger as logging_nemo
from nemo_bo.acquisition_functions.base_acq_function import AcquisitionFunction
from nemo_bo.acquisition_functions.nsga_improvement.problem_wrapper import NSGAProblemWrapper
from nemo_bo.acquisition_functions.nsga_improvement.unsga3 import UNSGA3
from nemo_bo.opt.constraints import ConstraintsList
from nemo_bo.opt.objectives import ObjectivesList
from nemo_bo.opt.samplers import LatinHyperCubeSampling, PoolBased, SampleGenerator
from nemo_bo.opt.variables import VariablesList

try:
    logging_nemo.logging_path
    logger = logging_nemo.logging_nemo_child(os.path.basename(__file__))
except AttributeError:
    logger = logging_nemo.logging_nemo_master(os.path.basename(__file__))


@dataclass
class NSGAImprovement(AcquisitionFunction):
    # def __init__(self, num_candidates):
    #     self.num_candidates = num_candidates
    num_candidates: int

    def generate_candidates(
        self,
        Y: np.ndarray,
        variables: VariablesList,
        objectives: ObjectivesList,
        sampler: Optional[SampleGenerator] = None,
        constraints: Optional[ConstraintsList] = None,
        **acq_func_kwargs,
    ) -> Tuple[np.ndarray, np.ndarray]:
        if isinstance(sampler, PoolBased):
            raise NotImplementedError(
                "The PoolBased sampler is incompatible with the NSGAImprovement acquisition function. Please select a different sampler if you wish to continue using this acquisition function. Alternatively, you may use the ExpectedImprovement acquisition function with the PoolBased sampler instead"
            )

        self.exploit_or_explore = acq_func_kwargs.get("exploit_or_explore", "exploit")
        self.pop_size = acq_func_kwargs.get("pop_size")
        self.generations = acq_func_kwargs.get("generations")

        Y = self.remove_nan_rows(Y)

        if self.pop_size is None:
            if self.num_candidates == 1:
                self.pop_size = 200
            elif self.num_candidates == 2:
                self.pop_size = 100
            elif self.num_candidates == 3:
                self.pop_size = 70
            elif self.num_candidates == 4:
                self.pop_size = 40

        if self.generations is None:
            self.generations = 30
            # if self.num_candidates == 1:
            #     self.generations = 50
            # elif self.num_candidates == 2:
            #     self.generations = 30
            # elif self.num_candidates == 3:
            #     self.generations = 30
            # elif self.num_candidates == 4:
            #     self.generations = 30

        if self.exploit_or_explore == "exploit":
            self.generate_pareto(variables, objectives, constraints)
            selected_X, selected_Y = self.hvi(self.res.X, self.res.F, Y, variables, objectives)

            self.reset = False

        if self.exploit_or_explore == "explore" or not self.hvi_search_pass:
            selected_X, selected_Y, total_Y_std = self.highest_uncertainty(variables, objectives, sampler, constraints)

            self.exploit_or_explore = "exploit"
            self.reset = True

        return selected_X, selected_Y

    def generate_pareto(
        self, variables: VariablesList, objectives: ObjectivesList, constraints: ConstraintsList
    ) -> Result:
        """
        Performs the NSGA2 multi-objective opt using the functions and predictor models provided

        Parameters
        ----------
        xl : numpy.ndarray or list for the lower bounds for the variables.
        xu : numpy.ndarray or list for the upper bounds for the variables.
        pop_size : int type for the population size in each iteration in NSGA-II
        generations : int type for the number of iterations in NSGA-II

        Returns
        -------
        res : pymoo.model.result.Result type object that contains all of the information for the predicted non-dominated solutions

        """

        xl = 0
        xu = 1

        ref_dirs = get_reference_directions("energy", objectives.n_obj, self.pop_size * 0.9, seed=1)

        logger.info(f"Performing U-NSGA3 opt")
        algorithm = UNSGA3(ref_dirs, pop_size=self.pop_size)
        problem = NSGAProblemWrapper(xl, xu, variables, objectives, constraints)
        termination = get_termination("n_gen", self.generations)

        # Performs the UNSGA3 opt
        self.res = minimize(problem, algorithm, termination, verbose=True)

        # Correct by * -1 for maximising objectives in NSGA2 opt
        for obj_index, obj in enumerate(objectives.max_bool_dict):
            if objectives.max_bool_dict[obj]:
                self.res.F[:, obj_index] *= -1

        # Un-normalise X values identified by NSGA
        self.res.X = (self.res.X * (np.array(variables.upper_bounds) - np.array(variables.lower_bounds))) + np.array(
            variables.lower_bounds
        )

        return self.res

    def hvi(
        self,
        X_nsga: np.ndarray,
        Y_nsga: np.ndarray,
        Y: np.ndarray,
        variables: VariablesList,
        objectives: ObjectivesList,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Identifies the candidate(s) from the non-dominated solutions from the NSGA2 opt that will produce the highest hypervolume or best improvement in d-metric when combined with the experimental data

        Parameters
        ----------
        X_nsga : numpy.ndarray type object with the variable values for all of the predicted non-dominated solutions
        Y_nsga : numpy.ndarray type object with the objective values for all of the predicted non-dominated solutions
        obj_minmax : 2D numpy.ndarray type of minimum and maximum values for each objective. The minimum value chosen should be slightly less than the expected experimental minimum to ensure accurate hypervolume calculation and pareto front identification
            Example: np.array([[min_yield, max_yield], [min_cost, max_cost]])
        number_of_suggestions : int type for the number of solutions to produce from the hypervolume sorting

        Returns
        -------
        selected_X : numpy.ndarray type object with the variable values for the non-dominated solution(s) that is predicted to produce best hypervolume improvement or best improvement in d-metric
        selected_Y : numpy.ndarray type object with the objectives values for the non-dominated solution(s) that is predicted to produce best hypervolume improvement or best improvement in d-metric

        """
        logger.info(f"Identifying conditions with the largest hypervolume")

        # Construct hvrefs and dmetric refs
        ref_point = self.build_ref_point(objectives.max_bool_dict)

        # finding min_max of full dataset
        sign_adjusted_Y = self.Y_norm_minmax_transform(Y, objectives.bounds, objectives.max_bool_dict)
        sign_adjusted_Y_nsga = self.Y_norm_minmax_transform(Y_nsga, objectives.bounds, objectives.max_bool_dict)

        # Creates a list of hypervolumes after each candidate(s) that successfully improves the hypervolume
        hv_list = []
        candidates_X_list = []
        candidates_Y_list = []
        X_nsga_combos = list(combinations(X_nsga, self.num_candidates))
        Y_nsga_combos = list(combinations(Y_nsga, self.num_candidates))
        sign_adjusted_Y_nsga_combos = list(combinations(sign_adjusted_Y_nsga, self.num_candidates))
        for x, y, candidates_Y in zip(X_nsga_combos, Y_nsga_combos, sign_adjusted_Y_nsga_combos):
            # Adds the candidate objective values to the experimental data set
            YappF = np.vstack([sign_adjusted_Y, np.array(candidates_Y)])
            # Identifies which data points are non-dominated
            pareto_mask_YappF = is_non_dominated(torch.tensor(YappF, dtype=torch.double))
            # Creates an numpy.ndarray containing the non-dominated points
            pareto_front_YappF = YappF[pareto_mask_YappF]

            # The following are checks to confirm that all new points in the candidates_Y set are on the pareto front
            candidate_in_pareto_counter = 0
            for a in candidates_Y:
                if any((a == b).all() for b in pareto_front_YappF):
                    candidate_in_pareto_counter += 1
                else:
                    break
            # So if all new points in the candidates_Y set are on the pareto front, then it will append the x and y and calculate the predicted hypervolume
            if candidate_in_pareto_counter == len(candidates_Y):
                candidates_X_list.append(x)
                candidates_Y_list.append(y)
                # Calculates the hypervolume
                hv = Hypervolume(ref_point=torch.tensor(ref_point, dtype=torch.double))
                hv_list.append(hv.compute(torch.tensor(pareto_front_YappF, dtype=torch.double)))

        # Selects the variable and objective values of the candidate(s) that produced the largest hypervolume
        # if hv_list is not empty
        if hv_list:
            logger.info("Conditions were successfully selected using hypervolume improvement")
            max_idx = hv_list.index(max(hv_list))
            selected_X = np.array(candidates_X_list[max_idx])
            selected_Y = np.array(candidates_Y_list[max_idx])
            self.hvi_search_pass = True

            if variables.num_cat_descriptor_var > 0:
                selected_X = variables.descriptor_to_name(selected_X)

        else:
            logger.info(
                "Failed to identify a set of conditions that would improve the hypervolume at the desired batch size"
            )
            selected_X = np.array([0])
            selected_Y = np.array([0])
            self.hvi_search_pass = False

        return selected_X, selected_Y

    def highest_uncertainty(
        self,
        variables: VariablesList,
        objectives: ObjectivesList,
        sampler: Optional[SampleGenerator] = None,
        constraints: Optional[ConstraintsList] = None,
        num_new_points: int = 200000,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Function to identify a set of input variables from the variable space that has the highest uncertainty for a predicted output value(s)

        Parameters
        ----------
        xl : 1D numpy.ndarray for the lower bounds for the variables.
        xu : 1D numpy.ndarray for the upper bounds for the variables.
        num_new_points : int type of the number of new sets of input variables to generate. Default: 200,000

        Returns
        -------
        self.uncertainty_dict : dict type containing the outputs relating to the uncertainty of the new data points generated

        """
        logger.info(f"Identifying conditions with the highest uncertainty")

        if sampler is None:
            self.sampler = LatinHyperCubeSampling(num_new_points)
        else:
            self.sampler = sampler

        X_new_points = self.sampler.generate_samples(variables, constraints)

        Y_new_points, Y_new_points_stddev = objectives.evaluate(X_new_points)

        # Min_max normalisation of Y standard deviation
        # I think this should be fine to minmax normalise like this
        for obj_index, _ in enumerate(objectives.bounds):
            Y_new_points_stddev[:, obj_index] = Y_new_points_stddev[:, obj_index] / (
                objectives.bounds[obj_index, 1] - objectives.bounds[obj_index, 0]
            )

        Y_new_points_stddev_sum = np.sum(Y_new_points_stddev, axis=1).reshape(-1, 1)
        sorting_indices = Y_new_points_stddev_sum[:, -1].argsort()

        X_new_points_sorted = X_new_points[sorting_indices][::-1]

        if variables.num_cat_var > 0:
            X_new_points = variables.categorical_transform(X_new_points).astype("float")
        X_new_points_norm_sorted = (
            (X_new_points[sorting_indices][::-1] - np.array(variables.lower_bounds))
            / (np.array(variables.upper_bounds) - np.array(variables.lower_bounds))
        )[sorting_indices][::-1]
        Y_new_points_sorted = Y_new_points[sorting_indices][::-1]
        Y_new_points_stddev_sum_sorted = Y_new_points_stddev_sum[sorting_indices][::-1]

        # 0 index for the highest uncertainty one at the top of the array
        chosen_indices = [0]
        # cosine similarity index of 1.0 for the highest uncertainty one at the top of the array
        cosine_simularity_list = [1.0]
        # Comparing the X_values for the highest uncertainty with every other entry down the sorted X_array
        for index, x in enumerate(X_new_points_norm_sorted):
            cosine_simularity_index = 1 - spatial.distance.cosine(X_new_points_norm_sorted[0], x)
            # Only passes if the cosine similarity index of a given x row is greater than 0.1 different from all of the x's in the list
            if all((((x - cosine_simularity_index) ** 2) ** 0.5) > 0.1 for x in cosine_simularity_list):
                cosine_simularity_list.append(cosine_simularity_index)
                chosen_indices.append(index)

            if len(chosen_indices) == self.num_candidates:
                break

        selected_X = X_new_points_sorted[chosen_indices]
        selected_Y = Y_new_points_sorted[chosen_indices].astype(np.float32)
        total_Y_std = Y_new_points_stddev_sum_sorted[chosen_indices].astype(np.float32)

        logger.info(f"Identified variables with the most uncertain output values")

        return selected_X, selected_Y, total_Y_std



import ast
import os
import pickle
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Dict, Optional, Tuple

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import torch
from botorch.utils.multi_objective.hypervolume import Hypervolume
from botorch.utils.multi_objective.pareto import is_non_dominated

import nemo_bo.utils.logger as logging_nemo
import nemo_bo.utils.plotter as plotter
from nemo_bo.acquisition_functions.base_acq_function import AcquisitionFunction
from nemo_bo.acquisition_functions.expected_improvement.expected_improvement import ExpectedImprovement
from nemo_bo.opt.benchmark import Benchmark
from nemo_bo.opt.constraints import ConstraintsList
from nemo_bo.opt.objectives import ObjectivesList, RegressionObjective
from nemo_bo.opt.samplers import LatinHyperCubeSampling, PoolBased, SampleGenerator
from nemo_bo.opt.variables import VariablesList

try:
    logging_nemo.logging_path
    logger = logging_nemo.logging_nemo_child(os.path.basename(__file__))
except AttributeError:
    logger = logging_nemo.logging_nemo_master(os.path.basename(__file__))


@dataclass
class Optimisation:
    variables: VariablesList
    objectives: ObjectivesList
    acquisition_func: AcquisitionFunction
    sampler: Optional[SampleGenerator | PoolBased] = None
    constraints: Optional[ConstraintsList] = None
    benchmark_func: Optional[Benchmark] = None

    def __post_init__(self):
        if isinstance(self.acquisition_func, ExpectedImprovement) and self.sampler is None:
            self.sampler = LatinHyperCubeSampling()

        self.optimisation_dict = {}

        self.X_columns = []
        for name, units in zip(self.variables.names, self.variables.units):
            if units == "":
                self.X_columns.append(name)
            else:
                self.X_columns.append(f"{name} ({units})")

        self.Y_columns = []
        for name, units in zip(self.objectives.names, self.objectives.units):
            if units == "":
                self.Y_columns.append(name)
            else:
                self.Y_columns.append(f"{name} ({units})")

    def run(
        self,
        X: np.ndarray,
        Y: np.ndarray,
        number_of_iterations: int = 200,
        test_ratio: float = 0.2,
        number_opt_iterations_completed: int = 0,
        model_run_counter: int = 0,
        model_run_counter_threshold: int = 5,
        plot_parity: bool = False,
        plot_pareto_hypervolume: bool = False,
        **kwargs,
    ) -> Dict[str, Dict[str, Any]]:
        self.X = X
        self.Y = Y

        for iteration in range(number_of_iterations):
            if model_run_counter % model_run_counter_threshold == 0:
                model_search_bool = True
                print("Will perform automated model and hyperparameter selection")
            else:
                model_search_bool = False
                print("Depending on predictor type, will perform model fitting with stored hyperparameters")

            if plot_pareto_hypervolume:
                # Plot pareto plot without predicted experiment values
                self.plot_scatter2d_opt(
                    Y=self.Y,
                    number_opt_iterations=number_opt_iterations_completed,
                    inc_predicted=False,
                )

                # Plot hypervolume plot without predicted experiment values
                self.plot_hypervolume(
                    Y=self.Y,
                    number_opt_iterations=number_opt_iterations_completed,
                    inc_predicted=False,
                )

            self.selected_X, self.selected_Y = self.find_candidates(
                self.X,
                self.Y,
                model_search_bool=model_search_bool,
                test_ratio=test_ratio,
                plot_parity=plot_parity,
                **kwargs,
            )

            suggested_XY = pd.DataFrame(
                np.hstack((self.selected_X, self.selected_Y)),
                columns=self.X_columns + self.Y_columns,
            )
            with pd.option_context("display.max_rows", None, "display.max_columns", None):
                print(suggested_XY)

            if not isinstance(self.sampler, PoolBased) and self.benchmark_func is None:
                if not os.path.exists(os.path.join(os.getcwd(), "Results")):
                    os.makedirs(os.path.join(os.getcwd(), "Results"))
                suggested_XY.to_excel(
                    os.path.join(
                        os.getcwd(),
                        "Results",
                        f"Iteration {iteration + 1} suggested candidates {datetime.now().strftime('%d-%m-%Y, %H-%M-%S')}.xlsx",
                    )
                )
                while True:
                    try:
                        actual_result = ast.literal_eval(
                            input(
                                f"Variables for the current opt iteration have been suggested (see print or 'Iteration {iteration + 1} suggested candidates.xlsx')\nPlease input the actual results in a python list"
                            )
                        )
                        assert isinstance(actual_result, list)
                    except AssertionError:
                        print(
                            "\nAssertionError: Please ensure that you have inputted the actual results in a python list\n"
                        )
                    try:
                        self.selected_Y = (
                            np.array(actual_result)
                            if self.acquisition_func.num_candidates > 1
                            else np.array(actual_result).reshape(1, -1)
                        )
                        assert (
                            self.selected_Y.shape[0] == self.acquisition_func.num_candidates
                            and self.selected_Y.shape[1] == self.objectives.n_obj
                        )
                        break
                    except AssertionError:
                        print(
                            "\nAssertionError: Please ensure that you have inputted the actual results in a python list with the correct dimensions for the number of candidates (x) and number of objectives (y)\n"
                        )

            self.X = np.vstack((self.X, self.selected_X))
            self.Y = np.vstack((self.Y, self.selected_Y))

            self.save(iteration, model_search_bool, model_run_counter)

            if self.acquisition_func.reset or model_search_bool:
                model_run_counter = 1
            else:
                model_run_counter += 1

            number_opt_iterations_completed += 1

        return self.optimisation_dict

    def find_candidates(self, X: np.ndarray, Y: np.ndarray, **kwargs) -> Tuple[np.ndarray, np.ndarray]:
        if any(isinstance(obj, RegressionObjective) for obj in self.objectives.objectives):
            plot_parity = kwargs.get("plot_parity", False)

            # Fit the ML regression models
            self.objectives.fit(X, Y, self.variables, **kwargs)

            if plot_parity:
                for obj in self.objectives.objectives:
                    obj.parity_plot()

        # Use the acquisition function
        selected_X, selected_Y = self.acquisition_func.generate_candidates(
            Y, self.variables, self.objectives, self.sampler, self.constraints, **kwargs
        )

        if isinstance(self.sampler, PoolBased):
            selected_Y = self.sampler.Y_and_update_pool(self.sampler.index)

        else:
            if self.benchmark_func is not None:
                if self.benchmark_func.fitted:
                    selected_Y, _ = self.benchmark_func.evaluate(self.selected_X)
                else:
                    raise AttributeError("Please firstly fit the benchmark function(s) and check its quality")

        return selected_X, selected_Y

    def save(self, iteration: int, model_search_bool: bool, model_run_counter: int) -> None:
        # Save excel for current iteration data
        if not os.path.exists(os.path.join(os.getcwd(), "Results")):
            os.makedirs(os.path.join(os.getcwd(), "Results"))

        model_info_list = [model_search_bool, model_run_counter]
        model_info_columns = [
            "Model and Hyperparameter Optimisation",
            "Model Run Counter",
        ]
        for obj in self.objectives.objectives:
            if isinstance(obj, RegressionObjective):
                model_info_list.append(obj.predictor_type)
                model_info_list.append(obj.model_params)
                model_info_list.append(obj.obj_function.performance_metrics["Train RMSE"])
                model_info_list.append(obj.obj_function.performance_metrics["Train r2"])
                model_info_columns.append(f"{obj.name} Model Type")
                model_info_columns.append(f"{obj.name} Model Hyperparameters")
                model_info_columns.append(f"{obj.name} Model Train RMSE")
                model_info_columns.append(f"{obj.name} Model Train r2")
                if obj.predictor_type in {
                    "nn_concrete",
                    "nn_ensemble",
                    "nn_bayesian",
                    "xgb",
                    "ngb",
                }:
                    model_info_list.append(obj.obj_function.performance_metrics["Validation RMSE"])
                    model_info_list.append(obj.obj_function.performance_metrics["Validation r2"])
                    model_info_columns.append(f"{obj.name} Model Validation RMSE")
                    model_info_columns.append(f"{obj.name} Model Validation r2")

        model_info_list.append(datetime.now().strftime("%d-%m-%Y, %H-%M-%S"))
        model_info_columns.append("Date and Time")
        model_info = np.array([model_info_list] * self.acquisition_func.num_candidates)

        self.optimisation_results = pd.DataFrame(np.hstack((self.X, self.Y)), columns=self.X_columns + self.Y_columns)
        if iteration == 0:
            self.optimisation_only = pd.DataFrame(
                np.hstack((self.selected_X, self.selected_Y, model_info)),
                columns=self.X_columns + self.Y_columns + model_info_columns,
            )
        else:
            self.optimisation_only = pd.concat(
                (
                    self.optimisation_only,
                    pd.DataFrame(
                        np.hstack((self.selected_X, self.selected_Y, model_info)),
                        columns=self.X_columns + self.Y_columns + model_info_columns,
                    ),
                ),
                axis=0,
                ignore_index=True,
            )

        with pd.ExcelWriter(
            os.path.join(
                os.getcwd(),
                "Results",
                f'Optimisation Results (after {iteration + 1} iteration(s)) {datetime.now().strftime("%d-%m-%Y, %H-%M-%S")}.xlsx',
            )
        ) as writer:
            self.optimisation_results.to_excel(writer, sheet_name="Training + Opt")
            self.optimisation_only.to_excel(writer, sheet_name="Only Opt")
            for obj in self.objectives.objectives:
                if isinstance(obj, RegressionObjective):
                    obj.obj_function.performance_metrics["Train Parity Data"].to_excel(
                        writer, sheet_name=f"{obj.name} Train Parity Data"
                    )
                    if obj.predictor_type in {
                        "nn_concrete",
                        "nn_ensemble",
                        "nn_bayesian",
                        "xgb",
                        "ngb",
                    }:
                        obj.obj_function.performance_metrics["Validation Parity Data"].to_excel(
                            writer, sheet_name=f"{obj.name} Validation Parity Data"
                        )

        # Saving all model parameters, predictor type, etc, in dictionaries per iteration as a pickle file
        self.optimisation_dict[f"Iteration {iteration + 1}"] = {}
        self.optimisation_dict[f"Iteration {iteration + 1}"]["Date and Time"] = datetime.now().strftime(
            "%d-%m-%Y, %H-%M-%S"
        )
        self.optimisation_dict[f"Iteration {iteration + 1}"][
            "Training + Optimisation Table"
        ] = self.optimisation_results
        self.optimisation_dict[f"Iteration {iteration + 1}"]["Optimisation Table"] = self.optimisation_only
        self.optimisation_dict[f"Iteration {iteration + 1}"]["Optimisation Selected X"] = self.selected_X
        self.optimisation_dict[f"Iteration {iteration + 1}"]["Optimisation Selected Y"] = self.selected_Y
        self.optimisation_dict[f"Iteration {iteration + 1}"][
            "Model and Hyperparameter Optimisation"
        ] = model_search_bool
        self.optimisation_dict[f"Iteration {iteration + 1}"]["Model Run Counter"] = model_run_counter
        for obj in self.objectives.objectives:
            if isinstance(obj, RegressionObjective):
                self.optimisation_dict[f"Iteration {iteration + 1}"][f"{obj.name} Model Type"] = obj.predictor_type
                self.optimisation_dict[f"Iteration {iteration + 1}"][
                    f"{obj.name} Model Hyperparameters"
                ] = obj.model_params
                self.optimisation_dict[f"Iteration {iteration + 1}"][
                    f"{obj.name} Model Performance Metrics"
                ] = obj.obj_function.performance_metrics

        with open(
            os.path.join(
                os.getcwd(),
                "Results",
                f"Optimisation Results (after {iteration + 1} iteration(s)) {datetime.now().strftime('%d-%m-%Y, %H-%M-%S')}.pickle",
            ),
            "wb",
        ) as handle:
            pickle.dump(self.optimisation_dict, handle, protocol=pickle.HIGHEST_PROTOCOL)

    def plot_scatter2d_opt(self, Y: np.ndarray, number_opt_iterations: int, inc_predicted: bool = False) -> None:

        sign_adjusted_Y = self.acquisition_func.Y_norm_minmax_transform(
            Y, self.objectives.bounds, self.objectives.max_bool_dict
        )

        # Calculates pareto front
        Y_to_expnum = torch.tensor(np.array(sign_adjusted_Y), dtype=torch.double)
        pareto_mask_to_expnum = is_non_dominated(Y_to_expnum)
        pareto_front = Y[pareto_mask_to_expnum]

        # To-do: Currently hard coded to sort for cost which can never be equal to 0
        cost_check = 0
        for obj_index, obj_name in enumerate(self.objectives.names):
            if "Cost" in obj_name:
                pareto_front_sorted = pareto_front[pareto_front[:, obj_index].argsort()]
                cost_check = 1
                break
        if cost_check == 0:
            pareto_front_sorted = pareto_front[pareto_front[:, 0].argsort()]

        if inc_predicted:
            train_experiments = Y[
                : (
                    Y.shape[0]
                    - (number_opt_iterations * self.acquisition_func.num_candidates)
                    - self.acquisition_func.num_candidates
                ),
                :,
            ]
            opt_experiments = Y[
                (
                    Y.shape[0]
                    - (number_opt_iterations * self.acquisition_func.num_candidates)
                    - self.acquisition_func.num_candidates
                ) :,
                :,
            ]
            opt_experiments = opt_experiments[: -self.acquisition_func.num_candidates, :]

            pred_experiments = Y[-self.acquisition_func.num_candidates :, :]
            scatter_data = [train_experiments, opt_experiments, pred_experiments]
            scatter_legend = [
                f"Training ({train_experiments.shape[0]} Experiments)",
                f"Optimisation ({opt_experiments.shape[0]} Experiments)",
                f"Predicted ({pred_experiments.shape[0]} Experiments)",
            ]

            title = "Pareto Plot (inc predicted experiments)"
            output = os.path.join(
                os.getcwd(),
                f'Pareto Plot (inc predicted experiments) {datetime.now().strftime("%d-%m-%Y, %H-%M-%S")}.png',
            )

        else:
            train_experiments = Y[
                : (Y.shape[0] - (number_opt_iterations * self.acquisition_func.num_candidates)),
                :,
            ]
            opt_experiments = Y[
                (Y.shape[0] - (number_opt_iterations * self.acquisition_func.num_candidates)) :,
                :,
            ]
            scatter_data = [train_experiments, opt_experiments]
            scatter_legend = [
                f"Training ({train_experiments.shape[0]} Experiments)",
                f"Optimisation ({opt_experiments.shape[0]} Experiments)",
            ]

            title = "Pareto Plot.png"
            output = os.path.join(
                os.getcwd(),
                f'Pareto Plot {datetime.now().strftime("%d-%m-%Y, %H-%M-%S")}.png',
            )

        plotter.plot(
            plot_dim="2D",
            scatter_data=scatter_data,
            line_data=pareto_front_sorted,
            scatter_legend=scatter_legend,
            line_legend="Pareto front",
            xlabel=f"{self.objectives.names[0]} {self.objectives.units[0]}",
            ylabel=f"{self.objectives.names[1]} {self.objectives.units[1]}",
            plottitle=f"{title}\n{self.objectives.names[0]} vs. {self.objectives.names[1]}",
            output_file=rf"{output}",
        )
        plt.close()

    def plot_hypervolume(self, Y: np.ndarray, number_opt_iterations: int, inc_predicted: bool = False) -> None:

        ref_point = self.acquisition_func.build_ref_point(self.objectives.max_bool_dict)

        sign_adjusted_Y = self.acquisition_func.Y_norm_minmax_transform(
            Y, self.objectives.bounds, self.objectives.max_bool_dict
        )
        # Recalculates the hypervolume for every additional row in the data
        hv_list = [0]
        for y_index in range(2, Y.shape[0] + 2):
            # Calculates pareto front
            obj_data_to_expnum = sign_adjusted_Y[:y_index, :]
            obj_data_to_expnum_np = np.array(obj_data_to_expnum)
            Y_to_expnum = torch.tensor(obj_data_to_expnum_np, dtype=torch.double)
            pareto_mask_to_expnum = is_non_dominated(Y_to_expnum)
            pareto_front_to_expnum = obj_data_to_expnum_np[pareto_mask_to_expnum]

            hv = Hypervolume(ref_point=torch.tensor(ref_point, dtype=torch.double))

            # Calculates the hypervolume
            hv_list.append(hv.compute(torch.tensor(pareto_front_to_expnum, dtype=torch.double)))

        if not inc_predicted:
            num_train_experiments = Y.shape[0] - (number_opt_iterations * self.acquisition_func.num_candidates)
            train_expt_list = [(x + 1) for x in list(range(num_train_experiments))]
            if number_opt_iterations > 0:
                opt_expt_list = [
                    train_expt_list[-1] + ((x + 1) * self.acquisition_func.num_candidates)
                    for x in list(range(number_opt_iterations))
                ]
                opt_num = opt_expt_list[-1] - num_train_experiments
                all_expt_list = train_expt_list + opt_expt_list
            else:
                opt_num = 0
                all_expt_list = train_expt_list

            title = f"Hypervolume\nNumber of training experiments: {num_train_experiments}\nNumber of opt experiments: {opt_num}"
            output_file = os.path.join(
                os.getcwd(),
                rf"Hypervolume vs. Experiment Number {datetime.now().strftime('%d-%m-%Y, %H-%M-%S')}.png",
            )

        else:
            num_train_experiments = (
                Y.shape[0]
                - (number_opt_iterations * self.acquisition_func.num_candidates)
                - self.acquisition_func.num_candidates
            )
            train_expt_list = [(x + 1) for x in list(range(num_train_experiments))]
            if number_opt_iterations > 0:
                opt_expt_list = [
                    train_expt_list[-1] + ((x + 1) * self.acquisition_func.num_candidates)
                    for x in list(range(number_opt_iterations))
                ]
                all_expt_list = train_expt_list + opt_expt_list + [Y.shape[0]]
                opt_num = opt_expt_list[-1] - num_train_experiments
            else:
                opt_num = 0
                all_expt_list = train_expt_list
            title = f"Hypervolume (including predicted hypervolume)\nNumber of training experiments: {num_train_experiments}\nNumber of opt experiments: {opt_num}\nNumber of predicted experiments: {self.acquisition_func.num_candidates}"
            output_file = os.path.join(
                os.getcwd(),
                rf"Hypervolume vs. Experiment Number (including predicted hypervolume) {datetime.now().strftime('%d-%m-%Y, %H-%M-%S')}.png",
            )

        hv_array = np.array([[expt_number, hv_list[expt_number]] for expt_number in all_expt_list])
        plotter.plot(
            plot_dim="2D",
            line_data=hv_array,
            line_legend="Hypervolume",
            xlabel=f"Experiment Number",
            ylabel=f"Normalised Hypervolume",
            plottitle=title,
            output_file=output_file,
        )
        plt.close()



from typing import List

from nemo_bo.models.base.base_model import Base_Model
from nemo_bo.models.gp import GPModel
from nemo_bo.models.ngb import NGBoostModel
from nemo_bo.models.nn_bayesian import NNBayesianModel
from nemo_bo.models.nn_concrete import NNConcreteDropoutModel
from nemo_bo.models.nn_ensemble import NNEnsembleModel
from nemo_bo.models.rf import RFModel
from nemo_bo.models.xgb import XGBoostModel


def create_predictor_list(predictor_type: List[str | Base_Model]) -> List[Base_Model]:
    if predictor_type is not None:
        if isinstance(predictor_type, list):
            predictor_type_list = []
            for predictor in predictor_type:
                if isinstance(predictor, str):
                    if predictor == "gp":
                        predictor_type_list.append(GPModel)
                    elif predictor == "nn_concrete":
                        predictor_type_list.append(NNConcreteDropoutModel)
                    elif predictor == "nn_ensemble":
                        predictor_type_list.append(NNEnsembleModel)
                    elif predictor == "nn_bayesian":
                        predictor_type_list.append(NNBayesianModel)
                    elif predictor == "xgb":
                        predictor_type_list.append(XGBoostModel)
                    elif predictor == "ngb":
                        predictor_type_list.append(NGBoostModel)
                    elif predictor == "rf":
                        predictor_type_list.append(RFModel)
                    else:
                        raise ValueError(
                            f'Predictor type "{predictor}" is not natively supported. Please provide the class of the external predictor that inherits from the BasePredictor class'
                        )
                else:
                    if isinstance(predictor, Base_Model):
                        predictor_type_list.append(predictor)
                    else:
                        raise ValueError(f'Predictor type "{predictor}" does not inherit from the BasePredictor class')
            return predictor_type_list

        elif isinstance(predictor_type, str):
            if predictor_type == "gp":
                return [GPModel]
            elif predictor_type == "nn_concrete":
                return [NNConcreteDropoutModel]
            elif predictor_type == "nn_ensemble":
                return [NNEnsembleModel]
            elif predictor_type == "nn_bayesian":
                return [NNBayesianModel]
            elif predictor_type == "xgb":
                return [XGBoostModel]
            elif predictor_type == "ngb":
                return [NGBoostModel]
            elif predictor_type == "rf":
                return [RFModel]
            else:
                raise ValueError(
                    f'Predictor type "{predictor_type}" is not natively supported. Please provide the class of the external predictor that inherits from the BasePredictor class'
                )
        else:
            raise TypeError(f'Please pass the predictor type "{predictor_type}" in a list or as a string')
    else:
        return [GPModel, NNConcreteDropoutModel, NNEnsembleModel, NNBayesianModel, XGBoostModel, NGBoostModel, RFModel]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nemo_bo',
 'nemo_bo.acquisition_functions',
 'nemo_bo.acquisition_functions.expected_improvement',
 'nemo_bo.acquisition_functions.nsga_improvement',
 'nemo_bo.models',
 'nemo_bo.models.base',
 'nemo_bo.opt',
 'nemo_bo.utils']

package_data = \
{'': ['*']}

install_requires = \
['botorch>=0.6.4,<0.7.0',
 'forestci>=0.5.1,<0.6.0',
 'gpytorch>=1.6.0,<2.0.0',
 'hyperopt>=0.2.7,<0.3.0',
 'matplotlib>=3.5.2,<4.0.0',
 'ngboost>=0.3.12,<0.4.0',
 'numpy>=1.21.6,<2.0.0',
 'pandas>=1.4.2,<2.0.0',
 'pymoo>=0.5.0,<0.6.0',
 'scikit-learn>=1.1.1,<2.0.0',
 'scipy>=1.8.1,<2.0.0',
 'six>=1.16.0,<2.0.0',
 'smt>=1.2.0,<2.0.0',
 'tensorflow-probability>=0.17.0,<0.18.0',
 'tensorflow>=2.9.1,<3.0.0',
 'torch>=1.11.0,<2.0.0',
 'xgboost-distribution>=0.2.4,<0.3.0']

setup_kwargs = {
    'name': 'nemo-bo',
    'version': '0.1.0',
    'description': 'Multi-objective optimization of chemical processes with automated machine learning model selection',
    'long_description': '#nemo\n\nMulti-objective optimization of chemical processes with automated machine learning model selection',
    'author': 'Simon Sung',
    'author_email': 'simon.sung06@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/sustainable-processes/NEMO',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import os
from typing import Any, List, Optional, Tuple

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.axes import Axes


def scatter_plot(
    ax: Axes,
    data: List[np.ndarray] | np.ndarray,
    legend: str | List[str],
    error: Optional[List[np.ndarray] | np.ndarray] = None,
) -> None:
    """
    Function to create the scatter plot Axes object

    :param ax: Matplotlib Axes object
    :param data: can be a 2D or 3D array or a list of 2D or 3D arrays for drawing multiple plots on to the same set of axes
    :param legend: a string or a list of strings for labelling the plots
    """

    # If the data provided is a array (i.e. not a list of arrays)
    if np.array(data)[0][0].ndim == 0:
        data_plot = np.array(data)
        if error is None:
            error = np.zeros(data_plot.shape[0])
        # If the array provided is 2D
        if data_plot.shape[1] == 2:
            # If legend information was provided
            if legend is not None:
                ax.errorbar(
                    data_plot[:, 0],
                    data_plot[:, 1],
                    yerr=error.flatten(),
                    fmt="o",
                    label=legend,
                    zorder=2,
                )
            else:
                ax.errorbar(
                    data_plot[:, 0],
                    data_plot[:, 1],
                    yerr=error.flatten(),
                    fmt="o",
                    label="Scatter plot",
                    zorder=2,
                )
        # Else if the array provided is 3D
        elif data_plot.shape[1] == 3:
            # If legend information was provided
            if legend is not None:
                ax.scatter(
                    data_plot[:, 0],
                    data_plot[:, 1],
                    data_plot[:, 2],
                    label=legend,
                    marker="o",
                    s=40,
                )
            else:
                ax.scatter(
                    data_plot[:, 0],
                    data_plot[:, 1],
                    data_plot[:, 2],
                    label="Scatter plot",
                    marker="o",
                    s=40,
                )
    # Else the data provided is a list of arrays
    else:
        # Iterate through the list of arrays and plot each one
        for plotnum in range(len(data)):
            data_plot = np.array(data[plotnum])
            if error is None:
                error_plot = np.zeros(data_plot.shape[0])
            else:
                error_plot = np.array(error[plotnum])
            # If the array provided is 2D
            if data_plot.shape[1] == 2:
                # If legend information was provided
                if legend is not None:
                    ax.errorbar(
                        data_plot[:, 0],
                        data_plot[:, 1],
                        yerr=error_plot.flatten(),
                        fmt="o",
                        label=legend[plotnum],
                        zorder=2,
                    )
                else:
                    ax.errorbar(
                        data_plot[:, 0],
                        data_plot[:, 1],
                        yerr=error_plot.flatten(),
                        fmt="o",
                        label=f"Scatter plot {plotnum + 1}",
                        zorder=2,
                    )
            # Else if the array provided is 3D
            elif data_plot.shape[1] == 3:
                # If legend information was provided
                if legend is not None:
                    ax.scatter(
                        data_plot[:, 0],
                        data_plot[:, 1],
                        data_plot[:, 2],
                        label=legend[plotnum],
                        marker="o",
                        s=40,
                    )
                else:
                    ax.scatter(
                        data_plot[:, 0],
                        data_plot[:, 1],
                        data_plot[:, 2],
                        label=f"Scatter plot {plotnum + 1}",
                        marker="o",
                        s=40,
                    )


def line_plot(data: List[np.ndarray] | np.ndarray, legend: str | List[str]) -> None:
    """
    Function to create the line plot object

    :param data: can be a 2D or 3D array or a list of 2D or 3D arrays for drawing multiple plots on to the same set of axes
    :param legend: a string or a list of strings for labelling the plots
    """
    # Starts the colour cycle for the plots from a different position (brown) to the default (blue)
    colorcycle = [
        "#8c564b",
        "#e377c2",
        "#7f7f7f",
        "#bcbd22",
        "#17becf",
        "#1f77b4",
        "#ff7f0e",
        "#2ca02c",
        "#d62728",
        "#9467bd",
    ]
    # If the data provided is a array (i.e. not a list of arrays)
    if np.array(data)[0][0].ndim == 0:
        data_plot = np.array(data)
        # If the array provided is 2D
        if data_plot.shape[1] == 2:
            # If legend information was provided
            if legend is not None:
                plt.plot(data_plot[:, 0], data_plot[:, 1], label=legend, color=colorcycle[0])
            else:
                plt.plot(
                    data_plot[:, 0],
                    data_plot[:, 1],
                    label="Line plot",
                    color=colorcycle[0],
                )
        # Else if the array provided is 3D
        elif data_plot.shape[1] == 3:
            # If legend information was provided
            if legend is not None:
                plt.plot(
                    data_plot[:, 0],
                    data_plot[:, 1],
                    data_plot[:, 2],
                    label=legend,
                    color=colorcycle[0],
                )
            else:
                plt.plot(
                    data_plot[:, 0],
                    data_plot[:, 1],
                    data_plot[:, 2],
                    label="Line plot",
                    color=colorcycle[0],
                )
    # Else the data provided is a list of arrays
    else:
        # Iterate through the list of arrays and plot each one
        for plotnum in range(len(data)):
            data_plot = np.array(data[plotnum])
            # If the array provided is 2D
            if data_plot.shape[1] == 2:
                # If legend information was provided
                if legend is not None:
                    plt.plot(
                        data_plot[:, 0],
                        data_plot[:, 1],
                        label=legend[plotnum],
                        color=colorcycle[plotnum],
                    )
                else:
                    plt.plot(
                        data_plot[:, 0],
                        data_plot[:, 1],
                        label=f"Line plot {plotnum + 1}",
                        color=colorcycle[plotnum],
                    )
            # Else if the array provided is 3D
            elif data_plot.shape[1] == 3:
                # If legend information was provided
                if legend is not None:
                    plt.plot(
                        data_plot[:, 0],
                        data_plot[:, 1],
                        data_plot[:, 2],
                        label=legend[plotnum],
                        color=colorcycle[0],
                    )
                else:
                    plt.plot(
                        data_plot[:, 0],
                        data_plot[:, 1],
                        data_plot[:, 2],
                        label=f"Line plot {plotnum + 1}",
                        color=colorcycle[0],
                    )


def surface_plot(ax: Axes, data: List[np.ndarray] | np.ndarray, legend: str | List[str]) -> None:
    """
    # Function to create the scatter plot Axes object

    :param ax: Matplotlib Axes object
    :param data: can be 3D array or a list of 3D arrays for drawing multiple plots on to the same set of axes
    :param legend: a string or a list of strings for labelling the plots
    """
    # Starts the colour cycle for the plots from a different position (brown) to the default (blue)
    colorcycle = [
        "#8c564b",
        "#e377c2",
        "#7f7f7f",
        "#bcbd22",
        "#17becf",
        "#1f77b4",
        "#ff7f0e",
        "#2ca02c",
        "#d62728",
        "#9467bd",
    ]
    # If the data provided is a array (i.e. not a list of arrays)
    if np.array(data)[0][0].ndim == 0:
        data_plot = np.array(data)
        # plot_trisurf breaks the plot when there are fewer than 3 points
        if data_plot.shape[0] >= 3:
            # If legend information was provided
            if legend is not None:
                surf = ax.plot_trisurf(
                    data_plot[:, 0],
                    data_plot[:, 1],
                    data_plot[:, 2],
                    label=legend,
                    alpha=0.3,
                    color=colorcycle[0],
                    antialiased=False,
                )
                surf._facecolors2d = surf._facecolor3d
                surf._edgecolors2d = surf._edgecolor3d
            else:
                surf = ax.plot_trisurf(
                    data_plot[:, 0],
                    data_plot[:, 1],
                    data_plot[:, 2],
                    label=f"Surface plot",
                    alpha=0.3,
                    color=colorcycle[0],
                    antialiased=False,
                )
                surf._facecolors2d = surf._facecolor3d
                surf._edgecolors2d = surf._edgecolor3d
    # Else the data provided is a list of arrays
    else:
        # Iterate through the list of arrays and plot each one
        for plotnum in range(len(data)):
            data_plot = np.array(data[plotnum])
            # plot_trisurf breaks the plot when there are fewer than 3 points
            if data_plot.shape[0] >= 3:
                # If legend information was provided
                if legend is not None:
                    surf = ax.plot_trisurf(
                        data_plot[:, 0],
                        data_plot[:, 1],
                        data_plot[:, 2],
                        label=legend[plotnum],
                        alpha=0.3,
                        color=colorcycle[plotnum],
                        antialiased=False,
                    )
                    surf._facecolors2d = surf._facecolor3d
                    surf._edgecolors2d = surf._edgecolor3d
                else:
                    surf = ax.plot_trisurf(
                        data_plot[:, 0],
                        data_plot[:, 1],
                        data_plot[:, 2],
                        label=f"Surface plot {plotnum + 1}",
                        alpha=0.3,
                        color=colorcycle[plotnum],
                        antialiased=False,
                    )
                    surf._facecolors2d = surf._facecolor3d
                    surf._edgecolors2d = surf._edgecolor3d


def plot3d_video(ax: Axes, output_file: str) -> None:
    """
    Creates a .mp4 video file of a rotating 3D Matplotlib plot using a series of images

    :param ax: Matplotlib Axes object
    :param output_file: a string of the output video location and file name. Do not need the video file extension in this string
    """
    angles = np.linspace(0, 360, 91)[:-1]  # Generates a list of 90 angles between 0 and 360 degrees

    # Creates a series of images at different angles and a list of image file names
    imagefiles_list = []
    print(f"Creating temporary image files")
    for i, angle in enumerate(angles):
        ax.view_init(elev=None, azim=angle)  # elevation set to None
        fname = f"tmpimg_{i}.jpeg"
        ax.figure.savefig(fname, dpi=400)  # dpi set to 400
        imagefiles_list.append(fname)

    # Uses mencoder.exe in the home directory to produce a .mp4 movie from a list of image files.
    command = f'mencoder "mf://{",".join(imagefiles_list)}" -mf fps={10} -o {output_file}.mp4 -ovc lavc -lavcopts vcodec=msmpeg4v2:vbitrate={8000}'  # fps=10, bitrate=8000
    os.system(command)

    # Deleted the temporary image files created above
    for f in imagefiles_list:
        os.remove(f)


def plot(
    plot_dim: str = "2D",
    scatter_data: Optional[List[np.ndarray | pd.DataFrame] | np.ndarray | pd.DataFrame] = None,
    error: Optional[List[np.ndarray] | np.ndarray] = None,
    line_data: Optional[List[np.ndarray | pd.DataFrame] | np.ndarray | pd.DataFrame] = None,
    surface_data: Optional[List[np.ndarray | pd.DataFrame] | np.ndarray | pd.DataFrame] = None,
    scatter_legend: Optional[str | List[str] | Tuple[str]] = None,
    line_legend: Optional[str | List[str]] = None,
    surface_legend: Optional[str | List[str]] = None,
    xlabel: str = "x",
    ylabel: str = "y",
    zlabel: str = "z",
    plottitle: str = "Title",
    output_file: str = "Plot",
    write_mp4: bool = False,
) -> None:
    """
    Function to start the plotting of a new 2D or 3D figure

    :param plot_dim: default is the string "2D". Refers to the number of dimensions for the desired plot
    :param scatter_data: default is None. The data used for the scatter plot can be either a 2D or 3D array or a list of 2D or 3D arrays
    :param line_data: default is None.  The data used for the line plot can be either a 2D or 3D array or a list of 2D or 3D arrays
    :param surface_data: default is None.  The data used for the surface plot can be either a 3D array or a list of 3D arrays
    :param scatter_legend: default is None. The string or list of strings used to label the new plot(s)
    :param line_legend: default is None. The string or list of strings used to label the new plot(s)
    :param surface_legend: default is None. The string or list of strings used to label the new plot(s)
    :param xlabel: default is "x". The string used to label the x-axis of the new figure
    :param ylabel: default is "y". The string used to label the y-axis of the new figure
    :param zlabel: default is "z". The string used to label the z-axis of the new figure
    :param plottitle: default is "Title". The string used as the title of the new figure
    :param output_file: default is "Plot". The string used to define the output file location of the .png and/or .mp4 of the new figure. Do not need the file extension (.png or .mp4) in this string
    :param write_mp4: default is False. When set to True, creates a .mp4 video file of a rotating 3D Matplotlib plot using a series of in-situ generated images
    """
    # Creates the Figure and Axes Matplotlib objects, with settings depending on if a 2D or 3D figure is requested
    if plot_dim == "2D":
        fig = plt.figure(figsize=(8, 8))
        ax = fig.add_subplot(1, 1, 1)
    elif plot_dim == "3D":
        fig = plt.figure()
        ax = fig.add_subplot(111, projection="3d")

    # Creates the scatter, line, or surface plot Axes objects respectively, depending on if the data for each type of plot was provided
    if scatter_data is not None:
        scatter_data = data_to_np(scatter_data)
        scatter_plot(ax, scatter_data, scatter_legend, error)
    if line_data is not None:
        line_data = data_to_np(line_data)
        line_plot(line_data, line_legend)
    if surface_data is not None:
        surface_data = data_to_np(surface_data)
        surface_plot(ax, surface_data, surface_legend)

    # Modifies the style settings for the figure, with settings depending on if a 2D or 3D figure is requested
    if plot_dim == "2D":
        ax.set_xlabel(xlabel, fontsize=12)
        ax.set_ylabel(ylabel, fontsize=12)
        ax.set_title(plottitle, fontsize=12)
        plt.legend(loc="best", fontsize=10)
    elif plot_dim == "3D":
        ax.set_xlabel(xlabel, fontsize=6)
        ax.set_ylabel(ylabel, fontsize=6)
        ax.set_zlabel(zlabel, fontsize=6)
        ax.tick_params(axis="both", which="major", labelsize=6)
        ax.set_title(plottitle, pad=24, fontsize=6)
        plt.legend(loc="best", fontsize=6)

    # Additional style settings for the figure
    ax.set_axisbelow(True)
    plt.grid(b=True, which="major", c="whitesmoke")

    # Writes a .png image file of the figure
    plt.savefig(f"{output_file}", dpi=400)

    # Writes a .mp4 video file of a rotating 3D Matplotlib plot
    if plot_dim == "3D":
        if write_mp4 == True:
            plot3d_video(ax, output_file)


def data_to_np(data: List[np.ndarray | pd.DataFrame] | np.ndarray | pd.DataFrame) -> List[np.ndarray] | np.ndarray:
    """
    # Function used to convert the list of DataFrames or numpy arrays into a dtype object numpy array of numpy arrays

    :param data: Any 2D or 3D array or list of 2D or 3D array
    :return: dtype object numpy array of numpy arrays
    """
    inputdata = data
    if isinstance(data, list):
        if isinstance(data[0], pd.DataFrame) or isinstance(data[0], np.ndarray):
            inputdata = np.zeros(len(data), dtype=np.dtype("object"))
            for x in range(len(inputdata)):
                inputdata[x] = np.array(data[x])

    return inputdata


def plot_bar(
    names: List[Any] | Tuple[Any] | np.ndarray,
    counts: List[Any] | Tuple[Any] | np.ndarray,
    xlabel: str = "x",
    ylabel: str = "y",
    legend: str = "Legend",
    plottitle: str = "Title",
    output_file: str = "Plot",
) -> None:
    fig, ax = plt.subplots(figsize=(8, 8))
    chart = ax.bar(names, counts, label=legend)
    ax.bar_label(chart, label_type="center")
    ax.set_xlabel(xlabel, fontsize=12)
    ax.set_ylabel(ylabel, fontsize=12)
    ax.set_title(plottitle, fontsize=12)

    plt.xticks(rotation=90)
    plt.legend()
    plt.tight_layout()

    plt.savefig(f"{output_file}", dpi=400)


def plot_stackedbar(
    names: List[Any] | Tuple[Any] | np.ndarray,
    bottom_counts: List[Any] | Tuple[Any] | np.ndarray,
    top_counts: List[Any] | Tuple[Any] | np.ndarray,
    xlabel: str = "x",
    ylabel: str = "y",
    bottom_legend: str = "Bottom legend",
    top_legend: str = "Top legend",
    plottitle: str = "Title",
    output_file: str = "Plot",
) -> None:
    fig, ax = plt.subplots(figsize=(8, 8))
    bar_bottom = ax.bar(names, bottom_counts, label=bottom_legend)
    bar_top = ax.bar(names, top_counts, bottom=bottom_counts, label=top_legend)
    ax.bar_label(bar_bottom, label_type="center")
    ax.bar_label(bar_top, label_type="center")
    ax.set_xlabel(xlabel, fontsize=12)
    ax.set_ylabel(ylabel, fontsize=12)
    ax.set_title(plottitle, fontsize=12)

    plt.xticks(rotation=90)
    plt.legend()
    plt.tight_layout()

    plt.savefig(f"{output_file}", dpi=400)


def plot_bar_horizontal(
    names: List[Any] | Tuple[Any] | np.ndarray,
    counts: List[Any] | Tuple[Any] | np.ndarray,
    xlabel: str = "x",
    ylabel: str = "y",
    legend: str = "Legend",
    plottitle: str = "Title",
    output_file: str = "Plot",
) -> None:
    fig, ax = plt.subplots(figsize=(8, 8))
    chart = ax.barh(names, counts, label=legend)
    ax.bar_label(chart, label_type="center")
    ax.set_xlabel(xlabel, fontsize=12)
    ax.set_ylabel(ylabel, fontsize=12)
    ax.set_title(plottitle, fontsize=12)

    plt.gca().invert_yaxis()
    plt.legend()
    plt.tight_layout()

    plt.savefig(f"{output_file}", dpi=400)


def plot_stackedbar_horizontal(
    names: List[Any] | Tuple[Any] | np.ndarray,
    left_counts: List[Any] | Tuple[Any] | np.ndarray,
    right_counts: List[Any] | Tuple[Any] | np.ndarray,
    xlabel: str = "x",
    ylabel: str = "y",
    left_legend: str = "Left legend",
    right_legend: str = "Right legend",
    plottitle: str = "Title",
    output_file: str = "Plot",
) -> None:
    fig, ax = plt.subplots(figsize=(8, 8))
    bar_left = ax.barh(names, left_counts, label=left_legend)
    bar_right = ax.barh(names, right_counts, left=left_counts, label=right_legend)
    ax.bar_label(bar_left, label_type="center")
    ax.bar_label(bar_right, label_type="center")
    ax.set_xlabel(xlabel, fontsize=12)
    ax.set_ylabel(ylabel, fontsize=12)
    ax.set_title(plottitle, fontsize=12)

    plt.gca().invert_yaxis()
    plt.legend()
    plt.tight_layout()

    plt.savefig(f"{output_file}", dpi=400)



# Adapted from:
# J. Blank and K. Deb, pymoo: Multi-Objective Optimization in Python, in IEEE Access, vol. 8, pp. 89497-89509, 2020, doi: 10.1109/ACCESS.2020.2990567; https://github.com/anyoptimization/pymoo

import numpy as np
from pymoo.core.algorithm import Algorithm
from pymoo.core.duplicate import DefaultDuplicateElimination, NoDuplicateElimination
from pymoo.core.initialization import Initialization
from pymoo.core.mating import Mating
from pymoo.core.population import Population
from pymoo.core.repair import NoRepair


class GeneticAlgorithm(Algorithm):
    def __init__(
        self,
        pop_size=None,
        sampling=None,
        selection=None,
        crossover=None,
        mutation=None,
        survival=None,
        n_offsprings=None,
        eliminate_duplicates=DefaultDuplicateElimination(),
        repair=None,
        mating=None,
        advance_after_initial_infill=False,
        **kwargs
    ):

        super().__init__(**kwargs)

        # the population size used
        self.pop_size = pop_size

        # whether the algorithm should be advanced after initialization of not
        self.advance_after_initial_infill = advance_after_initial_infill

        # the survival for the genetic algorithm
        self.survival = survival

        # number of offsprings to generate through recombination
        self.n_offsprings = n_offsprings

        # if the number of offspring is not set - equal to population size
        if self.n_offsprings is None:
            self.n_offsprings = pop_size

        # set the duplicate detection class - a boolean value chooses the default duplicate detection
        if isinstance(eliminate_duplicates, bool):
            if eliminate_duplicates:
                self.eliminate_duplicates = DefaultDuplicateElimination()
            else:
                self.eliminate_duplicates = NoDuplicateElimination()
        else:
            self.eliminate_duplicates = eliminate_duplicates

        # simply set the no repair object if it is None
        self.repair = repair if repair is not None else NoRepair()

        self.initialization = Initialization(
            sampling, repair=self.repair, eliminate_duplicates=self.eliminate_duplicates
        )

        if mating is None:
            mating = Mating(
                selection,
                crossover,
                mutation,
                repair=self.repair,
                eliminate_duplicates=self.eliminate_duplicates,
                n_max_iterations=100,
            )
        self.mating = mating

        # other run specific data updated whenever solve is called - to share them in all algorithms
        self.n_gen = None
        self.pop = None
        self.off = None

    def _initialize_infill(self):
        pop = self.initialization.do(self.problem, self.pop_size, algorithm=self)
        pop.set("n_gen", self.n_gen)

        if self.problem.variables.num_cat_var > 0:
            for individual_idx, individual in enumerate(pop):
                X_temp = (
                    individual.X
                    * (np.array(self.problem.variables.upper_bounds) - np.array(self.problem.variables.lower_bounds))
                ) + np.array(self.problem.variables.lower_bounds)
                X_temp = self.problem.variables.categorical_values_euc(X_temp.reshape(1, -1))
                pop[individual_idx].X = (X_temp.flatten() - np.array(self.problem.variables.lower_bounds)) / (
                    np.array(self.problem.variables.upper_bounds) - np.array(self.problem.variables.lower_bounds)
                )

        return pop

    def _initialize_advance(self, infills=None, **kwargs):
        if self.advance_after_initial_infill:
            self.pop = self.survival.do(self.problem, infills, n_survive=len(infills))

    def _infill(self):

        # do the mating using the current population
        off = self.mating.do(self.problem, self.pop, self.n_offsprings, algorithm=self)

        # if the mating could not generate any new offspring (duplicate elimination might make that happen)
        if len(off) == 0:
            self.termination.force_termination = True
            return

        # if not the desired number of offspring could be created
        elif len(off) < self.n_offsprings:
            if self.verbose:
                print("WARNING: Mating could not produce the required number of (unique) offsprings!")

        if self.problem.variables.num_cat_var > 0:
            for individual_idx, individual in enumerate(off):
                X_temp = (
                    individual.X
                    * (np.array(self.problem.variables.upper_bounds) - np.array(self.problem.variables.lower_bounds))
                ) + np.array(self.problem.variables.lower_bounds)
                X_temp = self.problem.variables.categorical_values_euc(X_temp.reshape(1, -1))
                off[individual_idx].X = (X_temp.flatten() - np.array(self.problem.variables.lower_bounds)) / (
                    np.array(self.problem.variables.upper_bounds) - np.array(self.problem.variables.lower_bounds)
                )

        return off

    def _advance(self, infills=None, **kwargs):

        # merge the offsprings with the current population
        if infills is not None:
            self.pop = Population.merge(self.pop, infills)

        # execute the survival to find the fittest solutions
        self.pop = self.survival.do(self.problem, self.pop, n_survive=self.pop_size, algorithm=self)



from dataclasses import dataclass
from typing import Dict, List

import numpy as np
from pymoo.core.problem import Problem

from nemo_bo.opt.constraints import ConstraintsList
from nemo_bo.opt.objectives import ObjectivesList
from nemo_bo.opt.variables import VariablesList


@dataclass
class NSGAProblemWrapper(Problem):
    """
    Class to be instantiated when defining a problem for the minimise function in NSGA2
    """

    def __init__(
        self,
        xl: int | float | List[int | float],
        xu: int | float | List[int | float],
        variables: VariablesList,
        objectives: ObjectivesList,
        constraints: ConstraintsList,
    ):
        """
        Parameters
        ----------
        xl : numpy.ndarray or list or int type for the lower bounds for the variables. If integer all lower bounds are equal (inherited from the Problem class)
        xu : numpy.ndarray or list or int type for the upper bounds for the variables. If integer all upper bounds are equal (inherited from the Problem class)
        n_var : int type for the number of variables (inherited from the Problem class)
        n_obj : int type for the number of objectives (inherited from the Problem class)
        n_constr : int type for the number of constraints (inherited from the Problem class)
        model : Model type object used for evaluating black box objectives in the _evaluate function
        objname_list : list type of the names of the objectives
        obj_max_bool : dict type with the objective names as keys and the values are a boolean, where True correlates to the objective being maximising
        inputchempath : str type of the file path for the input chemicals excel file for the opt
        reactscale : int type for the volume of pump A solution (in mL) used for each reaction
        """
        self.variables = variables
        self.objectives = objectives
        self.constraints = constraints

        n_constr = self.constraints.n_constr if constraints is not None else 0

        super().__init__(
            n_var=self.variables.n_var,
            n_obj=self.objectives.n_obj,
            n_constr=n_constr,
            xl=xl,
            xu=xu,
        )

    def _evaluate(self, X: np.ndarray, out: Dict[str, np.ndarray], **kwargs):
        """
        Function called at line 385 of the pymoo/model/problem.py to calculate the objective values and re-insert them back into NSGA2

        Parameters
        ----------
        X : 1D or 2D numpy.ndarray type of un_transformed input variables
        out : dict type containing the outputs from the NSGA2 for each individual
        kwargs : dict type containing the reaction parameters used for evaluating the objectives for Cost and EcoScore. Initially constructed at line 86 and line 123 of the pymoo/algorithms/genetic_algorithm.py

        Returns
        -------
        out: Appends the out dict with the key "F" containing F, which is a numpy.ndarray with the shape corresponding to the number of individuals defined in X and the number of objectives

        """

        X = (X * (np.array(self.variables.upper_bounds) - np.array(self.variables.lower_bounds))) + np.array(
            self.variables.lower_bounds
        )

        if self.variables.num_cat_descriptor_var > 0:
            F, _ = self.objectives.evaluate(self.variables.descriptor_to_name(X))
        else:
            F, _ = self.objectives.evaluate(X)

        for obj_index, obj in enumerate(self.objectives.max_bool_dict):
            if self.objectives.max_bool_dict[obj]:
                F[:, obj_index] *= -1
        out["F"] = F

        if self.n_constr > 0:
            G = np.zeros((X.shape[0], self.n_constr))
            for index, constr in enumerate(self.constraints.constraints):
                G[:, index] = constr.evaluate_pymoo_constraint(X)
            out["G"] = G



# Copied from:
# J. Blank and K. Deb, pymoo: Multi-Objective Optimization in Python, in IEEE Access, vol. 8, pp. 89497-89509, 2020, doi: 10.1109/ACCESS.2020.2990567; https://github.com/anyoptimization/pymoo

import numpy as np
from pymoo.operators.selection.tournament import TournamentSelection, compare

from nemo_bo.acquisition_functions.nsga_improvement.nsga3 import NSGA3


# =========================================================================================================
# Implementation
# =========================================================================================================


def comp_by_rank_and_ref_line_dist(pop, P, **kwargs):
    S = np.full(P.shape[0], np.nan)

    for i in range(P.shape[0]):
        a, b = P[i, 0], P[i, 1]

        # if at least one solution is infeasible
        if pop[a].CV > 0.0 or pop[b].CV > 0.0:
            S[i] = compare(
                a,
                pop[a].CV,
                b,
                pop[b].CV,
                method="smaller_is_better",
                return_random_if_equal=True,
            )

        # both solutions are feasible
        else:

            # if in the same niche select by rank
            if pop[a].get("niche") == pop[b].get("niche"):

                if pop[a].get("rank") != pop[b].get("rank"):
                    S[i] = compare(
                        a,
                        pop[a].get("rank"),
                        b,
                        pop[b].get("rank"),
                        method="smaller_is_better",
                    )

                else:
                    S[i] = compare(
                        a,
                        pop[a].get("dist_to_niche"),
                        b,
                        pop[b].get("dist_to_niche"),
                        method="smaller_is_better",
                    )

        if np.isnan(S[i]):
            S[i] = np.random.choice([a, b])

    return S[:, None].astype(int)


class UNSGA3(NSGA3):
    def __init__(self, ref_dirs, selection=TournamentSelection(func_comp=comp_by_rank_and_ref_line_dist), **kwargs):
        super().__init__(ref_dirs, selection=selection, **kwargs)



from __future__ import annotations

import copy
import os
from datetime import datetime
from typing import TYPE_CHECKING, Any, Dict, Optional

import matplotlib.pyplot as plt
import pandas as pd
from sklearn.metrics import mean_squared_error
from sklearn.model_selection import KFold, train_test_split

import nemo_bo.utils.logger as logging_nemo
import nemo_bo.utils.perf_metrics as pm
import nemo_bo.utils.plotter as plotter
from nemo_bo.utils.data_proc import *

if TYPE_CHECKING:
    from nemo_bo.opt.objectives import RegressionObjective
    from nemo_bo.opt.variables import VariablesList

try:
    logging_nemo.logging_path
    logger = logging_nemo.logging_nemo_child(os.path.basename(__file__))
except AttributeError:
    logger = logging_nemo.logging_nemo_master(os.path.basename(__file__))


class Base_Model:
    """
    Class to instantiate for fitting GP models
    """

    def __init__(
        self, variables: VariablesList, objective: RegressionObjective, always_hyperparam_opt: bool = True, **kwargs
    ):
        self.variables = variables
        self.input_variables = copy.copy(variables)
        self.objective = objective
        self.input_objective = copy.copy(objective)
        self.always_hyperparam_opt = always_hyperparam_opt
        self.default_X_transform_type = None
        self.default_Y_transform_type = None
        self.name = None

    def new_instance(self, cls, **params):
        return cls(
            copy.copy(self.input_variables), copy.copy(self.input_objective), self.always_hyperparam_opt, **params
        )

    def fit(
        self,
        X: np.ndarray,
        Y: np.ndarray,
        test_ratio: float = 0.2,
        sort_before_split: Optional[bool] = None,
        X_transform: bool = True,
        **params,
    ) -> None:
        pass

    def fit_model(self, *args, **kwargs):
        pass

    def predict(self, X: np.ndarray, X_transform: bool = True) -> Tuple[np.ndarray, np.ndarray]:
        pass

    def cv(self, *args, **kwargs) -> Dict[str, Any]:
        pass

    def hyperparam_opt(self, *args, **kwargs) -> Dict[str, Any]:
        pass

    def default_params(self, *args, **kwargs) -> Dict[str, Any]:
        pass

    def cv_train_test(
        self, X: np.ndarray, Y: np.ndarray, model_params: Dict[str, Any], test_ratio: float = 0.2, **kwargs
    ) -> Dict[str, Any]:

        self.model_params = model_params

        cv_results = {}
        test_rmse_list = []
        test_r2_list = []
        fold_number = 0

        kf = KFold(n_splits=int(1 / test_ratio))

        # Split the data
        X_temp = copy.copy(X)
        Y_temp = copy.copy(Y)

        remainder = X.shape[0] % int(1 / test_ratio)
        if remainder > 0:
            X_temp = X_temp[: (-1 * remainder)]
            Y_temp = Y_temp[: (-1 * remainder)]

        for train_index, test_index in kf.split(X_temp):
            fold_number += 1
            X_train, X_test = X_temp[train_index], X_temp[test_index]
            Y_train, Y_test = (
                Y_temp[train_index].astype("float"),
                Y_temp[test_index].astype("float"),
            )

            # Create copy for the model
            model = self.new_instance(self.__class__, **kwargs)

            # Transform the data
            model.X_train, model.Y_train, model.X_test, model.Y_test = (
                X_train,
                Y_train,
                X_test,
                Y_test,
            )
            if model.variables.num_cat_var > 0:
                model.X_train = model.variables.categorical_transform(model.X_train).astype("float")
                model.X_test = model.variables.categorical_transform(model.X_test).astype("float")

            model.X_train, model.Y_train = model.transform_by_predictor_type(model.X_train, Y=model.Y_train)
            model.X_test, model.Y_test = model.transform_only_by_predictor_type(model.X_test, Y=model.Y_test)

            model.fit_model(self.model_params)

            # Make predictions and calculate model performance
            model.Y_train_pred, model.Y_train_pred_stddev = model.predict(X_train)
            model.Y_test_pred, model.Y_test_pred_stddev = model.predict(X_test)
            model.performance_metrics = pm.all_performance_metrics_train_val(
                Y_train, model.Y_train_pred, Y_test, model.Y_test_pred
            )

            cv_results[f"Fold {fold_number}"] = {
                "Performance Metrics": model.performance_metrics,
                "Model": model.model,
            }
            test_rmse_list.append(model.performance_metrics["Validation RMSE"])
            test_r2_list.append(model.performance_metrics["Validation r2"])

        cv_results["Mean Test RMSE"] = np.mean(test_rmse_list)
        cv_results["Mean Test r2"] = np.mean(test_r2_list)

        logger.info(
            f"Completed {self.name} CV (k = {int(1/test_ratio)}) model checking | Mean Test Loss: {cv_results['Mean Test RMSE']}"
        )

        return cv_results

    def cv_train_val_test(
        self,
        X: np.ndarray,
        Y: np.ndarray,
        model_params: Dict[str, Any],
        test_ratio: float = 0.2,
        sort_before_split: bool = True,
        **kwargs,
    ) -> Dict[str, Any]:

        self.model_params = model_params

        cv_results = {}
        test_rmse_list = []
        test_r2_list = []
        test_pred_list = []
        test_error_list = []
        fold_number = 0

        kf = KFold(n_splits=int(1 / test_ratio))

        if sort_before_split:
            (
                X_train_val,
                X_test,
                Y_train_val,
                Y_test,
            ) = sort_train_test_split_shuffle(X, Y, test_ratio=test_ratio, seed=1)
        else:
            X_train_val, X_test, Y_train_val, Y_test = train_test_split(X, Y, test_size=test_ratio, random_state=1)
        Y_test = Y_test.astype("float")

        for train_index, test_index in kf.split(X_train_val):
            fold_number += 1
            X_train, X_val = X_train_val[train_index], X_train_val[test_index]
            Y_train, Y_val = (
                Y_train_val[train_index].astype("float"),
                Y_train_val[test_index].astype("float"),
            )

            # Create copy for the model
            model = self.new_instance(self.__class__, **kwargs)

            # Transform the data
            (model.X_train, model.Y_train, model.X_val, model.Y_val, model.X_test, model.Y_test,) = (
                X_train,
                Y_train,
                X_val,
                Y_val,
                X_test,
                Y_test,
            )
            if model.variables.num_cat_var > 0:
                model.X_train = model.variables.categorical_transform(model.X_train).astype("float")
                model.X_val = model.variables.categorical_transform(model.X_val).astype("float")
                model.X_test = model.variables.categorical_transform(model.X_test).astype("float")

            model.X_train, model.Y_train = model.transform_by_predictor_type(model.X_train, model.Y_train)
            model.X_val, model.Y_val = model.transform_only_by_predictor_type(model.X_val, model.Y_val)
            (
                model.X_test,
                model.Y_test,
            ) = model.transform_only_by_predictor_type(model.X_test, model.Y_test)

            model.fit_model(self.model_params)

            # Make predictions and calculate model performance
            model.Y_train_pred, model.Y_train_pred_stddev = model.predict(X_train)
            model.Y_val_pred, model.Y_val_pred_stddev = model.predict(X_val)
            model.Y_test_pred, model.Y_test_pred_stddev = model.predict(X_test)
            model.performance_metrics = pm.all_performance_metrics_train_val_test(
                Y_train,
                model.Y_train_pred,
                Y_val,
                model.Y_val_pred,
                Y_test,
                model.Y_test_pred,
            )
            model.Y_test_error = 1.96 * model.Y_test_pred_stddev

            cv_results[f"Fold {fold_number}"] = {
                "Performance Metrics": model.performance_metrics,
                "Model": model.model,
            }
            test_rmse_list.append(model.performance_metrics["Test RMSE"])
            test_r2_list.append(model.performance_metrics["Test r2"])
            test_pred_list.append(model.Y_test_pred.flatten())
            test_error_list.append(model.Y_test_error.flatten())

        cv_results["Mean Test RMSE"] = np.mean(test_rmse_list)
        cv_results["Mean Test r2"] = np.mean(test_r2_list)
        cv_results["Mean Y Test Parity Data"] = pm.paritydata(Y_test, np.mean(test_pred_list, axis=0))
        cv_results["Mean Y Test Error (CI 95%)"] = np.mean(test_error_list, axis=0)

        logger.info(
            f"Completed {self.name} CV (k = {int(1/test_ratio)}) model checking | Mean Test Loss: {cv_results['Mean Test RMSE']}"
        )

        return cv_results

    def non_cv_train_test(
        self,
        X: np.ndarray,
        Y: np.ndarray,
        model_params: Dict[str, Any],
        test_ratio: float = 0.2,
        sort_before_split: bool = True,
        **kwargs,
    ) -> float:

        self.model_params = model_params

        # Split the dataset
        if sort_before_split:
            (
                X_train,
                X_test,
                Y_train,
                Y_test,
            ) = sort_train_test_split_shuffle(X, Y, test_ratio=test_ratio, seed=1)
        else:
            X_train, X_test, Y_train, Y_test = train_test_split(X, Y, test_size=test_ratio, random_state=1)

        # Create copy for the model
        model = self.new_instance(self.__class__, **kwargs)

        # Transform the data
        model.X_train, model.Y_train, model.X_test, model.Y_test = (
            X_train,
            Y_train,
            X_test,
            Y_test,
        )
        if model.variables.num_cat_var > 0:
            model.X_train = model.variables.categorical_transform(model.X_train).astype("float")
            model.X_test = model.variables.categorical_transform(model.X_test).astype("float")

        model.X_train, model.Y_train = model.transform_by_predictor_type(model.X_train, Y=model.Y_train)
        model.X_test, model.Y_test = model.transform_only_by_predictor_type(model.X_test, Y=model.Y_test)

        model.fit_model(self.model_params)

        # Make predictions and calculate model performance
        model.Y_train_pred, model.Y_train_pred_stddev = model.predict(X_train)
        model.Y_test_pred, model.Y_test_pred_stddev = model.predict(X_test)
        model.performance_metrics = pm.all_performance_metrics_train_val(
            Y_train, model.Y_train_pred, Y_test, model.Y_test_pred
        )

        return model.performance_metrics["Test RMSE"]

    def non_cv_train_val_test(
        self,
        X: np.ndarray,
        Y: np.ndarray,
        model_params: Dict[str, Any],
        test_ratio: float = 0.2,
        sort_before_split: bool = True,
        **kwargs,
    ) -> float:
        self.model_params = model_params

        # Split the dataset
        if sort_before_split:
            (
                X_train,
                X_val,
                X_test,
                Y_train,
                Y_val,
                Y_test,
            ) = sort_train_val_test_split_shuffle(X, Y, test_ratio=test_ratio, seed=1)
        else:
            X_train_val, X_test, Y_train_val, Y_test = train_test_split(X, Y, test_size=test_ratio, random_state=1)
            X_train, X_val, Y_train, Y_val = train_test_split(
                X_train_val, Y_train_val, test_size=test_ratio, random_state=1
            )

        # Create copy for the model
        model = self.new_instance(self.__class__, **kwargs)

        # Transform the data
        if model.variables.num_cat_var > 0:
            model.X_train = model.variables.categorical_transform(X_train).astype("float")
            model.X_val = model.variables.categorical_transform(X_val).astype("float")
            model.X_test = model.variables.categorical_transform(X_test).astype("float")

        model.X_train, model.Y_train = model.transform_by_predictor_type(model.X_train, Y=Y_train)
        model.X_val, model.Y_val = model.transform_only_by_predictor_type(model.X_val, Y=Y_val)
        model.X_test, model.Y_test = model.transform_only_by_predictor_type(model.X_test, Y=Y_test)

        model.fit_model(self.model_params)

        # Make predictions and calculate model performance
        model.Y_train_pred, model.Y_train_pred_stddev = model.predict(X_train)
        model.Y_val_pred, model.Y_val_pred_stddev = model.predict(X_val)
        model.Y_test_pred, model.Y_test_pred_stddev = model.predict(X_test)
        model.performance_metrics = pm.all_performance_metrics_train_val_test(
            Y_train, model.Y_train_pred, Y_val, model.Y_val_pred, Y_test, model.Y_test_pred
        )

        return model.performance_metrics["Test RMSE"]

    def y_scrambling_cv(
        self,
        X: np.ndarray,
        Y: np.ndarray,
        model_params: Dict[str, Any],
        test_ratio: float = 0.2,
        plot_parity=True,
        inc_error_bars: bool = False,
        **kwargs,
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """
        Compares the model fitting and accuracy of test data using cross validation for the original data set and a Y-scrambled data set

        Parameters
        ----------
        predictor_type : str or list of str type for the type of model to use. Default is None. When None, all model types in NEMO will be tested
            "gp" : Gaussian process models
            "nn_concrete" : Neural networks with automatic tuning of the dropout probability
            "nn_ensemble" : Deep Ensembles of neural networks
            "nn_bayesian" : Bayesian neural network from learning weight distributions
            "xgb" : XGBoost for probabilistic prediction
            "ngb" : NGBoost algorithm with natural gradient boosting for probabilistic prediction
        model_params : dict type containing the keyword hyperparameter arguments to use for model fitting
        plot_parity : bool type for generating the parity plot for the model fitting performance
        inc_error_bars : bool type for if error bars corresponding to a 95% confidence interval should drawn in the parity plot

        Returns
        -------
        cv_results : dict type containing the performance metrics from the fitted models using the original data set, with emphasis on the test data Results
        cv_results_y_scrambling : dict type containing the performance metrics from the fitted models using the Y-scrambled data set, with emphasis on the test data Results

        """

        dirname_y_scrambling = os.path.join(os.getcwd(), "ML Models", f"{self.name}", "Y-Scrambling")
        if not os.path.exists(dirname_y_scrambling):
            os.makedirs(dirname_y_scrambling)

        cv_results = self.cv_train_val_test(X, Y, model_params, test_ratio=test_ratio, **kwargs)

        np.random.seed(0)
        np.random.shuffle(Y)

        cv_results_y_scrambling = self.cv_train_val_test(X, Y, model_params, test_ratio=test_ratio, **kwargs)

        if plot_parity:
            # Produces a 2D scatter plot showing the parity plot data points, and the x = y line as a line plot
            original_paritydata = cv_results["Mean Y Test Parity Data"]
            y_scrambling_paritydata = cv_results_y_scrambling["Mean Y Test Parity Data"]

            original_error = cv_results["Mean Y Test Error (CI 95%)"]
            y_scrambling_error = cv_results_y_scrambling["Mean Y Test Error (CI 95%)"]

            original_rmse = cv_results["Mean Test RMSE"]
            y_scrambling_rmse = cv_results_y_scrambling["Mean Test RMSE"]
            original_r2 = cv_results["Mean Test r2"]
            y_scrambling_r2 = cv_results_y_scrambling["Mean Test r2"]

            scatter_parity_data = [original_paritydata, y_scrambling_paritydata]
            error = [original_error, y_scrambling_error]
            max_parity = np.vstack([original_paritydata, y_scrambling_paritydata]).max()
            min_parity = np.vstack([original_paritydata, y_scrambling_paritydata]).min()
            # Array for the x = y line in the parity plot
            x_equals_y = [
                [min_parity, min_parity],
                [max_parity, max_parity],
            ]
            scatter_legend = (
                f"Original ({original_paritydata.shape[0]} Test Data Points)",
                f"Y-Scrambling ({y_scrambling_paritydata.shape[0]} Test Data Points)",
            )
            plot_title = f"Original and Y-Scrambling Test Data Parity Plot\nOriginal Test RMSE = {round(original_rmse.astype('float'), 2)}, Original Test r2 = {round(original_r2, 2)}\nY-Scrambling Test RMSE = {round(y_scrambling_rmse.astype('float'), 2)}, Y-Scrambling Test r2 = {round(y_scrambling_r2, 2)}"
            output_file = os.path.join(
                dirname_y_scrambling,
                f"Y-Scrambling Test Data Parity Plot, {datetime.now().strftime('%d-%m-%Y, %H-%M-%S')}",
            )
            if inc_error_bars:
                plotter.plot(
                    plot_dim="2D",
                    scatter_data=scatter_parity_data,
                    error=error,
                    line_data=x_equals_y,
                    scatter_legend=scatter_legend,
                    line_legend="x = y",
                    xlabel=f"Actual",
                    ylabel=f"Predicted",
                    plottitle=plot_title,
                    output_file=output_file,
                )
            else:
                plotter.plot(
                    plot_dim="2D",
                    scatter_data=scatter_parity_data,
                    # error=error,
                    line_data=x_equals_y,
                    scatter_legend=scatter_legend,
                    line_legend="x = y",
                    xlabel=f"Actual",
                    ylabel=f"Predicted",
                    plottitle=plot_title,
                    output_file=output_file,
                )
            plt.close()

        logger.info(
            f"Completed Y-scrambling CV tests | Mean Test Loss: {cv_results['Mean Test RMSE']} | Y-scrambling Mean Test Loss: {cv_results_y_scrambling['Mean Test RMSE']}"
        )

        return cv_results, cv_results_y_scrambling

    def permutation_feature_importance(
        self, X: np.ndarray, Y: np.ndarray, repeats: int = 50
    ) -> Tuple[pd.Series, pd.Series]:
        logger.info(f"Performing permutation feature importance calculation for {self.name} model")

        Y_pred_base, _ = self.predict(X)
        mse_score_base = mean_squared_error(Y, Y_pred_base)

        results = []
        for i in range(X.shape[1]):
            temp_i_store = copy.copy(X[:, i])

            mse_score_i_shuffled_list = []
            for repeat in range(repeats):
                np.random.shuffle(X[:, i])
                Y_pred_i_shuffled, _ = self.predict(X)
                mse_score_i_shuffled = mean_squared_error(Y, Y_pred_i_shuffled)
                mse_score_i_shuffled_list.append(mse_score_i_shuffled)
                if repeat % 10 == 0:
                    print(
                        f"Variable {i + 1}, Repeat {repeat + 1}, Permutation MSE: {mse_score_i_shuffled}, Difference to Base MSE: {mse_score_i_shuffled - mse_score_base}"
                    )
            X[:, i] = copy.copy(temp_i_store)
            results.append(np.array(mse_score_i_shuffled_list))

        results = np.array(results).T
        results_diff = results - mse_score_base
        feature_importance_mean = np.mean(results_diff, axis=0)
        feature_importance_mean = feature_importance_mean / np.sum(feature_importance_mean)
        feature_importance_stddev = np.std(results_diff, axis=0)

        feature_importance_mean = pd.Series(feature_importance_mean, index=self.variables.names)
        feature_importance_stddev = pd.Series(feature_importance_stddev, index=self.variables.names)

        print(f"Permutation feature importance: {np.around(feature_importance_mean, decimals=3)}")

        logger.info(f"Completed permutation feature importance calculation")

        return feature_importance_mean, feature_importance_stddev

    def permutation_feature_importance_continuous(
        self, X: np.ndarray, Y: np.ndarray, repeats: int = 50
    ) -> Tuple[pd.Series, pd.Series]:
        logger.info(f"Performing permutation feature importance calculation for {self.name} model")

        Y_pred_base, _ = self.predict(X)
        mse_score_base = mean_squared_error(Y, Y_pred_base)

        # Convert categorical variables names to descriptors
        if self.variables.num_cat_var > 0:
            X = self.variables.categorical_transform(X)

        # Transform the continuous variables and descriptors
        X = self.transform_only_by_predictor_type(X)

        results = []
        for i in range(X.shape[1]):
            temp_i_store = copy.copy(X[:, i])

            mse_score_i_shuffled_list = []
            for repeat in range(repeats):
                np.random.shuffle(X[:, i])

                Y_pred_i_shuffled, _ = self.predict(X, X_transform=False)

                mse_score_i_shuffled = mean_squared_error(Y, Y_pred_i_shuffled.flatten())
                mse_score_i_shuffled_list.append(mse_score_i_shuffled)
                if repeat % 10 == 0:
                    print(
                        f"Variable {i + 1}, Repeat {repeat + 1}, Permutation MSE: {mse_score_i_shuffled}, Difference to Base MSE: {mse_score_i_shuffled - mse_score_base}"
                    )
            X[:, i] = copy.copy(temp_i_store)
            results.append(np.array(mse_score_i_shuffled_list))

        results = np.array(results).T
        results_diff = results - mse_score_base
        feature_importance_mean = np.mean(results_diff, axis=0)
        feature_importance_mean = feature_importance_mean / np.sum(feature_importance_mean)
        feature_importance_stddev = np.std(results_diff, axis=0)

        feature_importance_mean = pd.Series(feature_importance_mean, index=self.variables.continuous_var_names)
        feature_importance_stddev = pd.Series(feature_importance_stddev, index=self.variables.continuous_var_names)

        print(f"Permutation feature importance:\n{np.around(feature_importance_mean, decimals=3)}")

        logger.info(f"Completed permutation feature importance calculation")

        return feature_importance_mean, feature_importance_stddev

    def partial_dependence_plot(self, X: np.ndarray) -> Dict[str, pd.DataFrame]:
        logger.info(f"Creating partial dependence plots for {self.name} model")
        path = os.path.join(os.getcwd(), "ML Models", f"{self.name}", "Partial Dependence Plots")
        if not os.path.exists(path):
            os.makedirs(path)

        if self.variables.num_cat_var > 0:
            X = self.variables.categorical_transform(X)

        X = self.transform_only_by_predictor_type(X)

        X_copy = copy.copy(X)

        pdp_dict = {}
        for index, col in enumerate(X.T):
            Y_pred_pdp = []
            for value in col:
                X_copy[:, index] = value

                Y_pred, _ = self.predict(X_copy, X_transform=False)
                Y_pred_pdp.append([value, Y_pred.mean()])

            Y_pred_pdp_np = np.array(Y_pred_pdp)
            pdp_dict[self.variables.continuous_var_names[index]] = pd.DataFrame(
                Y_pred_pdp_np,
                columns=[
                    f"{self.variables.continuous_var_names[index]}",
                    f"{self.objective.name}",
                ],
            )
            X_copy[:, index] = col

            plotter.plot(
                plot_dim="2D",
                scatter_data=Y_pred_pdp_np,
                scatter_legend=f"Number of Grid Points = {Y_pred_pdp_np.shape[0]}",
                xlabel=f"{self.variables.continuous_var_names[index]}",
                ylabel=f"{self.objective.name} ({self.objective.units})",
                plottitle=f"Partial Dependence Plot for {self.objective.name} against {self.variables.continuous_var_names[index]}",
                output_file=os.path.join(
                    path,
                    rf"Partial Dependence Plot for {self.objective.name} against {self.variables.continuous_var_names[index]}, {datetime.now().strftime('%d-%m-%Y, %H-%M-%S')}",
                ),
            )
            plt.close("all")

        logger.info(f"Completed partial dependence plot calculations")

        return pdp_dict

    def transform_by_predictor_type(
        self, X: np.ndarray, Y: Optional[np.ndarray] = None
    ) -> np.ndarray | Tuple[np.ndarray, np.ndarray]:
        X = self.variables.transform(X, self.default_X_transform_type)

        if Y is not None:
            if self.objective.transformation_type is None:
                self.objective.transformation_type = self.default_Y_transform_type
            Y = self.objective.transform(Y)
            return X.astype("float"), Y.astype("float")
        else:
            return X.astype("float")

    def transform_only_by_predictor_type(
        self, X: np.ndarray, Y: Optional[np.ndarray] = None
    ) -> np.ndarray | Tuple[np.ndarray, np.ndarray]:
        X = self.variables.transform_only(X)

        if Y is not None:
            Y = self.objective.transform_only(Y)
            return X.astype("float"), Y.astype("float")
        else:
            return X.astype("float")
